// -*- Mode:C++ -*-
//Header:
//File: AuElman5L.h
//Title: Autistic Elman Net with 5 layers
//Author: NODA, Itsuki
//Date: 2000/01/05
//

//ModifyHistory:
// 2000/01/05: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_AuElman5L_h_
#define _Nnn_AuElman5L_h_
////////////////////////////////////////////////////////////////////////

#include "BpNetBase.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // Autistic Elman 5 Layer

   class AuElman5L : public BpNetBase {

      public:
	 //--------------------------------------------------
	 // constants

	 //------------------------------
	 // feedback repeat count

	 static const UInt fb_max = 1 ;

	 //------------------------------
	 // amount of adjust context

	 static const Val adjustRate = 1.0;

	 //------------------------------
	 // for save/load/scan

	 static char* const _sformTermHead = "AuElman5L" ; 
	 virtual char* sformTermHead() const {
	    return _sformTermHead ;
	 } ;

	 static char * const nameContext = "context" ;
	 static char * const nameInter1 = "inter1" ;
	 static char * const nameHidden = "hidden" ;
	 static char * const nameInter2 = "inter2" ;
	 static char * const nameOutput = "output" ;

	 //--------------------------------------------------
	 // member
      public:
	 BpLayer * context ;
	 BpLayer * inter1 ;
	 BpLayer * hidden ;
	 BpLayer * inter2 ;
	 BpLayer * output ;

	 BpLLink * con_int1 ;
	 BpLLink * int1_hid ;
	 BpLLink * hid_int2 ;
	 BpLLink * int2_out ;

	 ValVector * ctxvec ;

	 //--------------------------------------------------
	 // constructor
      public:
	 AuElman5L() : ctxvec() { init() ; } ;

	 //--------------------------------------------------
	 // alloc new memory
     public:
	 void alloc(const SString & nm,
		    const UInt & int1sz,
		    const UInt & hidsz,
		    const UInt & int2sz,
		    const UInt & outsz,
		    const Val & tempval = NNN_DFLT_TEMP,
		    const Val & epsval = NNN_DFLT_EPS,
		    const Val & alphaval = NNN_DFLT_ALPHA) {
	    init(nm,tempval,epsval,alphaval) ;
	    assignLayerLLink(int1sz,hidsz,int2sz,outsz) ;
	 } ;
	 
	 //--------------------------------------------------
	 // assignLayer (l-s-l-s-l)
      public:
	 void assignLayerLLink(const UInt & int1sz,
			       const UInt & hidsz, 
			       const UInt & int2sz,
			       const UInt & outsz) {
	    context = newLayer(nameContext,hidsz,BpLayer::FT_linear) ;
	    inter1 = newLayer(nameInter1,int1sz,BpLayer::FT_sigmoid) ;
	    hidden = newLayer(nameHidden,hidsz,BpLayer::FT_linear) ;
	    inter2 = newLayer(nameInter2,int2sz,BpLayer::FT_sigmoid) ;
	    output = newLayer(nameOutput,outsz,BpLayer::FT_linear) ;
	    
	    con_int1 = newLLink(context,inter1) ;
	    int1_hid = newLLink(inter1,hidden);
	    hid_int2 = newLLink(hidden,inter2) ;
	    int2_out = newLLink(inter2,output) ;
	 } ;

	 //--------------------------------------------------
	 // propFore
      public:
	 void propFore(Bool firstp = True) {
	    if(firstp) {
	       if(isNull(ctxvec)) ctxvec = new ValVector(hidden->n) ;

	       hidden->getOut(ctxvec) ;
	       context->setOut(ctxvec) ;
	    }
	    
	    inter1->propFore(temp) ;
	    hidden->propFore(temp) ;
	    inter2->propFore(temp) ;
	    output->propFore(temp) ;
	 } ;
	    
	 //--------------------------------------------------
	 // propBack
      public:
	 void propBack(const ValVector & tchvec) {
	    output->propBack(tchvec,temp) ;
	    inter2->propBack(temp) ;
	    hidden->propBack(temp) ;
	    inter1->propBack(temp) ;
	    context->propBack(temp) ;
	 } ;

	 //--------------------------------------------------
	 // adjust Context
      public:
	 void adjustContext(const Val & eps) {
	    context->adjustOutputByError(eps) ;
	 } ;

	 //--------------------------------------------------
	 // learn
      public:
	 Flt learn() {
	    Val r = 0.0 ;
	    r += inter1->learn(eps,alpha) ;
	    r += hidden->learn(eps,alpha) ;
	    r += inter2->learn(eps,alpha) ;
	    r += output->learn(eps,alpha) ;

	    r += con_int1->learn(eps,alpha) ;
	    r += int1_hid->learn(eps,alpha) ;
	    r += hid_int2->learn(eps,alpha) ;
	    r += int2_out->learn(eps,alpha) ;
	    return r ;
	 } ;

	 //--------------------------------------------------
	 // run
      public:
	 void run() { return propFore(True) ; } ;
	 void run(const ValVector & tchvec) { 
	    for(UInt i = 0 ; i < fb_max ; i++) {
	       if(i==0) propFore(True) ; 
	       else     propFore(False) ;
	       propBack(tchvec) ;
	       adjustContext(adjustRate) ;
	    } ;
	    propFore(False) ;
	 } ;
	 void runLearn(const ValVector & tchvec) {
	    for(UInt i = 0 ; i < fb_max ; i++) {
	       if(i==0) propFore(True) ; 
	       else     propFore(False) ;
	       propBack(tchvec) ;
	       learn() ;
	       adjustContext(adjustRate) ;
	    } ;
	    propFore(False) ;
	    propBack(tchvec) ;
	    learn() ;
	 } ;

	 //--------------------------------------------------
	 // get sum of square error
      public:
	 Flt sumError() const { return output->sumError() ; } ;

	 //--------------------------------------------------
	 // get output
      public:
	 void getNetOutput(ValVector & outvec) const { 
	    output->getNetOutput(outvec) ;
	 } ;

	 //--------------------------------------------------
	 // scan
      public:
	 virtual void scan(StrForm & sform) {
	    BpNetBase::scan(sform) ;

	    context = layer(nameContext) ;
	    inter1 = layer(nameInter1) ;
	    hidden = layer(nameHidden) ;
	    inter2 = layer(nameInter2) ;
	    output = layer(nameOutput) ;

	    con_int1 = link(context,inter1) ;
	    int1_hid = link(inter1,hidden) ;
	    hid_int2 = link(hidden,inter2) ;
	    int2_out = link(inter2,output) ;
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif


