// -*- Mode:C++ -*-
//Header:
//File: BpLayer.h
//Author: NODA, Itsuki
//Date: 2000/01/12
//

//ModifyHistory:
// 2000/01/12: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_BpLayer_h_
#define _Nnn_BpLayer_h_
////////////////////////////////////////////////////////////////////////

#include "n3btype.h"
#include "n3utility.h"
#include "StrForm.h"
#include "ValVector.h"
#include "BpUnitBase.h"
#include "BpLinkBase.h"

namespace Nnn {

   using namespace Itk ;

   class BpLLink ;

   //======================================================================
   // BpLayer

   class BpLayer : public WithDescriber {

	 //--------------------------------------------------
	 // labels for save/load/scan
      public:
	 static char * const sformTermHead = "BpLayer" ; 
	 static char * const sformNameLbl = "name" ; 
	 static char * const sformSizeLbl = "n" ; 
	 static char * const sformUnitLbl = "unit" ; 
	 static char * const sformInitLbl = "init" ; 
	 static char * const sformFuncTypeLbl = "ftype" ; 
	 static char * const sformLrnRateLbl = "lrate" ; 

	 //--------------------------------------------------
	 // typedef
      public:
	 typedef StackT<BpLLink*> LLinkList ;

	 //--------------------------------------------------
	 // Function Type
      public:
	 enum FuncType { FT_linear, FT_sigmoid, FT_bigmoid } ;
	 static const char * ftypeName(FuncType type) {
	    switch(type) {
	       case FT_linear : return "linear" ;
	       case FT_sigmoid : return "sigmoid" ;
	       case FT_bigmoid : return "bigmoid" ;
	    }
	    return ITK_NULLPTR ;
	 } ;
	 static FuncType ftypeByName(const SString & name) {
	    if(name == "linear") return FT_linear ;
	    if(name == "sigmoid") return FT_sigmoid ;
	    if(name == "bigmoid") return FT_bigmoid ;
	    throw 0 ;
	 } ;

	 //--------------------------------------------------
	 // member
      public:
	 SubString name ;
	 Size n ;
	 FuncType ftype ;
	 BpUnitBase * _unit ;
	 Val lrate ;
	 
	 // Val temp ;	// means to use network's temp
	 // Val eps ;	// learning step
	 // Val alpha ;	// ineatia
	 
	 StrForm * initInfo ;

	 LLinkList from ;
	 LLinkList to ;

	 //--------------------------------------------------
	 // constructor
      public:
	 BpLayer() { init(SString()) ; } ;
	 BpLayer(const SubString & nm) { init(nm) ; } ;
	 BpLayer(const SubString & nm, Size size) { init(nm,size) ; } ;
	 BpLayer(const SubString & nm, Size size, FuncType ft) { 
	    init(nm,size,ft) ; 
	 } ;
	 BpLayer(const SubString & nm, Size size, FuncType ft, 
		 BpUnitBase * u) {
	    init(nm,size,ft,u) ;
	 } ;

	 //--------------------------------------------------
	 // initialize
      public:
	 void init(const SubString & nm, Size size = 0, 
		   FuncType ft = FT_sigmoid,
		   BpUnitBase * u = ITK_NULLPTR) {
	    name.copy(nm,True) ;
	    n = size ;
	    ftype = ft ;
	    if(isNull(u)) {
	       if(n > 0) {
		  _unit = new BpUnitBase[n] ;
	       } else {
		  _unit = ITK_NULLPTR ;
	       } 
	    } else {
	       _unit = u ;
	    } 
	    lrate = 1.0 ;
	 } ;

	 //--------------------------------------------------
	 // resize
      public:
	 Bool resize(Size newsize) {
	    Size oldsize = n ;
	    n = newsize ;
	    if(oldsize > newsize) return False ;
	    
	    if(!isNull(_unit)) delete _unit ;
	    _unit = new BpUnitBase[n] ;
	    return True ;
	 } ;

	 //--------------------------------------------------
	 // unit(i) : access to each unit
      public:
	 const BpUnitBase & unit(const Index & i) const { return _unit[i] ; } ;

	 BpUnitBase & unit(const Index & i) { return _unit[i] ; } ;

	 //--------------------------------------------------
	 // access to each vector
      public:
	 void setInp(const ValVector & inpvec) ;
	 void setInp(const ValVector * inpvec) { setInp(*inpvec) ; } ;
	 void getInp(ValVector & inpvec) const ;
	 void getInp(ValVector * inpvec) const { getInp(*inpvec) ; } ;

	 void setOut(const ValVector & outvec) ;
	 void setOut(const ValVector * outvec) { setOut(*outvec) ; } ;
	 void getOut(ValVector & outvec) const ;
	 void getOut(ValVector * outvec) const { getOut(*outvec) ; } ;

	 void setThr(const ValVector & thrvec) ;
	 void setThr(const ValVector * thrvec) { setThr(*thrvec) ; } ;
	 void getThr(ValVector & thrvec) const ;
	 void getThr(ValVector * thrvec) const { getThr(*thrvec) ; } ;

	 void setErr(const ValVector & errvec) ;
	 void setErr(const ValVector * errvec) { setErr(*errvec) ; } ;
	 void getErr(ValVector & errvec) const ;
	 void getErr(ValVector * errvec) const { getErr(*errvec) ; } ;

	 void setDlt(const ValVector & dltvec) ;
	 void setDlt(const ValVector * dltvec) { setDlt(*dltvec) ; } ;
	 void getDlt(ValVector & dltvec) const ;
	 void getDlt(ValVector * dltvec) const { getDlt(*dltvec) ; } ;

	 void setDh(const ValVector & dhvec) ;
	 void setDh(const ValVector * dhvec) { setDh(*dhvec) ; } ;
	 void getDh(ValVector & dhvec) const ;
	 void getDh(ValVector * dhvec) const { getDh(*dhvec) ; } ;

	 //--------------------------------------------------
	 // forward propagation
	 //	propFore(inpvec,temp) : for input layer
	 //	propFore(temp) : for hidden & output layer
      public:
	 void propFore(const ValVector & inpvec, const Val & temp) ;
	 void propFore(const Val & temp) ;

	 void setNetInput(const ValVector & ivec) ;
	 void getNetOutput(ValVector & ovec) ;

	 void propForeLinear(const Val & temp) ;
	 void propForeSigmoid(const Val & temp) ;
	 void propForeBigmoid(const Val & temp) ;

	 void clearInput(const Val & value = 0.0) ;
	 void clearInputByThreshold() ;

	 void collectInput(const Bool clearp = True) ;

	 void activateByLinear(const Val & temp) ;
	 //void activateByLinear() { return activateByLinear(temp) ; } ;

	 void activateBySigmoid(const Val & temp) ;
	 //void activateBySigmoid() { return activateBySigmoid(temp) ; } ;

	 void activateByBigmoid(const Val & temp) ;
	 //void activateByBigmoid() { return activateByBigmoid(temp) ; } ;

	 //--------------------------------------------------
	 // backward propagation
	 //	propBack(tchvec,temp): for output layer
	 //	propBack(temp) : for hidden layer
      public:
	 void propBack(const ValVector & tchvec, const Val & temp) ;
	 void propBack(const Val & temp) ;

	 void propBackLinear(const ValVector & tchvec, 
			     const Val & temp) ;
	 void propBackSigmoid(const ValVector & tchvec,
			      const Val & temp) ;
	 void propBackBigmoid(const ValVector & tchvec,
			      const Val & temp) ;

	 void propBackLinear(const Val & temp) ;
	 void propBackSigmoid(const Val & temp) ;
	 void propBackBigmoid(const Val & temp) ;

	 void clearError() ;

	 void calcNetError(const ValVector & tchvec) ;
	 void collectDelta() ;
	 void calcDeltaByLinear(const Val & temp) ;
	 void calcDeltaBySigmoid(const Val & temp) ;
	 void calcDeltaByBigmoid(const Val & temp) ;

	 //--------------------------------------------------
	 // adjust output by error value
      public:
	 void adjustOutputByError(const Val & eps) ;

	 //--------------------------------------------------
	 // calc sum of error

	 Val sumError() const ;

	 //--------------------------------------------------
	 // learn
      public:
	 Val learn(const Val & eps, const Val & alpha) ;

	 //--------------------------------------------------
	 // add link
      public:
	 Bool addFromLink(BpLLink * lnk) ;
	 Bool addToLink(BpLLink * lnk) ;

	 //--------------------------------------------------
	 // randomize threshold
      public:
	 void randomizeThreshold(const Val & minval, const Val & maxval) ;

	 //--------------------------------------------------
	 // save 
      public:
	 virtual void save(ostream & ostr,
			   const Bool & newlinep = True,
			   const SubString & indent = SubString("")) const ;

	 //--------------------------------------------------
	 // load
      public:
	 virtual void load(istream & istr) {
	    StrForm * sform = StrForm::scan(istr) ;
	    scan(*sform) ;
	    delete sform ;
	 } ;

      public:
	 virtual void scan(StrForm & sform) ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#BpLayer[" 
		 << name
		 << "|size=" << n
		 << ", ftype=" << ftypeName(ftype) 
		 << "]" ;

	    if(detailp) {
	       ostr << endl ;
	       ostr << "  " << "unit= (" ;
	       for(Index i = 0 ; i < n ; i++) {
		  if(i > 0) { ostr << "," << endl << "         " ; } 
		  ostr << unit(i) ;
	       }
	       ostr << ")" << endl ;
	       ostr << "  " << "from= " ; from.describe(ostr) ;
	       ostr << "  " << "to= " ; to.describe(ostr) ;
	    } 
	 } ;

	 //--------------------------------------------------
	 // iterator
      public:
	 typedef BpUnitBase * Iterator ;
	 typedef const BpUnitBase * C_Iterator ;
	 
	 Iterator begin() { return _unit ; } ;
	 C_Iterator begin() const { return _unit ; } ;
	 Iterator end() { return &(_unit[n]) ; } ;
	 C_Iterator end() const { return &(_unit[n]) ; } ;
   } ;

   //======================================================================
   // BpLayerTable

   class BpLayerTable : public NameTableT<BpLayer *> {
      public:
	 Entry * assign(BpLayer * lyr) {
	    return NameTable::assign(lyr->name, lyr) ;
	 } ;
	 Entry * assign(BpLayer & lyr) {  return assign(&lyr) ; } ;
   } ;

} ;

#include "BpLLink.h"

////////////////////////////////////////////////////////////////////////
#endif
