// -*- Mode:C++ -*-
//Header:
//File: Elman5L.h
//Title: Elman Net with 5 sigmoid layers
//Author: NODA, Itsuki
//Date: 2000/01/05
//

//ModifyHistory:
// 2000/01/05: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_Elman5L_h_
#define _Nnn_Elman5L_h_
////////////////////////////////////////////////////////////////////////

#include "BpNetBase.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // Elman 5 Layer

   class Elman5L : public BpNetBase {

      public:
	 //--------------------------------------------------
	 // constants

	 static char* const _sformTermHead = "Elman5L" ; 
	 virtual char* sformTermHead() const {
	    return _sformTermHead ;
	 } ;

	 static char * const nameContext = "context" ;
	 static char * const nameInput = "input" ;
	 static char * const nameInter1 = "inter1" ;
	 static char * const nameHidden = "hidden" ;
	 static char * const nameInter2 = "inter2" ;
	 static char * const nameOutput = "output" ;

	 //--------------------------------------------------
	 // member
      public:
	 BpLayer * context ;
	 BpLayer * input ;
	 BpLayer * inter1 ;
	 BpLayer * hidden ;
	 BpLayer * inter2 ;
	 BpLayer * output ;

	 BpLLink * con_int1 ;
	 BpLLink * inp_int1 ;
	 BpLLink * int1_hid ;
	 BpLLink * hid_int2 ;
	 BpLLink * int2_out ;

	 ValVector * ctxvec ;

	 //--------------------------------------------------
	 // constructor
      public:
	 Elman5L() : ctxvec() { init() ; } ;

	 //--------------------------------------------------
	 // alloc new memory
     public:
	 void alloc(const SString & nm,
		    const UInt & inpsz,
		    const UInt & int1sz,
		    const UInt & hidsz,
		    const UInt & int2sz,
		    const UInt & outsz,
		    const Val & tempval = NNN_DFLT_TEMP,
		    const Val & epsval = NNN_DFLT_EPS,
		    const Val & alphaval = NNN_DFLT_ALPHA) {
	    init(nm,tempval,epsval,alphaval) ;
	    assignLayerLLink(inpsz,int1sz,hidsz,int2sz,outsz) ;
	 } ;
	 
	 //--------------------------------------------------
	 // assignLayer (l-s-l-s-l)
      public:
	 void assignLayerLLink(const UInt & inpsz, 
			       const UInt & int1sz,
			       const UInt & hidsz, 
			       const UInt & int2sz,
			       const UInt & outsz) {
	    context = newLayer(nameContext,hidsz,BpLayer::FT_linear) ;
	    input = newLayer(nameInput,inpsz,BpLayer::FT_linear) ;
	    inter1 = newLayer(nameInter1,int1sz,BpLayer::FT_sigmoid) ;
	    hidden = newLayer(nameHidden,hidsz,BpLayer::FT_linear) ;
	    inter2 = newLayer(nameInter2,int2sz,BpLayer::FT_sigmoid) ;
	    output = newLayer(nameOutput,outsz,BpLayer::FT_linear) ;
	    
	    con_int1 = newLLink(context,inter1) ;
	    inp_int1 = newLLink(input,inter1) ;
	    int1_hid = newLLink(inter1,hidden);
	    hid_int2 = newLLink(hidden,inter2) ;
	    int2_out = newLLink(inter2,output) ;
	 } ;

	 //--------------------------------------------------
	 // propFore
      public:
	 void propFore(const ValVector & inpvec) {
	    if(isNull(ctxvec)) ctxvec = new ValVector(hidden->n) ;

	    hidden->getOut(ctxvec) ;
	    context->setOut(ctxvec) ;
	    input->propFore(inpvec,temp) ;
	    
	    inter1->propFore(temp) ;
	    hidden->propFore(temp) ;
	    inter2->propFore(temp) ;
	    output->propFore(temp) ;
	 } ;
	    
	 //--------------------------------------------------
	 // propBack
      public:
	 void propBack(const ValVector & tchvec) {
	    output->propBack(tchvec,temp) ;
	    inter2->propBack(temp) ;
	    hidden->propBack(temp) ;
	    inter1->propBack(temp) ;
	    // lyrInput->propBack(inpvec,temp) ;
	 } ;

	 //--------------------------------------------------
	 // learn
      public:
	 Flt learn() {
	    Val r = 0.0 ;
	    r += inter1->learn(eps,alpha) ;
	    r += hidden->learn(eps,alpha) ;
	    r += inter2->learn(eps,alpha) ;
	    r += output->learn(eps,alpha) ;

	    r += con_int1->learn(eps,alpha) ;
	    r += inp_int1->learn(eps,alpha) ;
	    r += int1_hid->learn(eps,alpha) ;
	    r += hid_int2->learn(eps,alpha) ;
	    r += int2_out->learn(eps,alpha) ;
	    return r ;
	 } ;

	 //--------------------------------------------------
	 // run
      public:
	 void run(const ValVector & inpvec) { return propFore(inpvec) ; } ;
	 void runLearn(const ValVector & inpvec,
		       const ValVector & tchvec) {
	    propFore(inpvec) ;
	    propBack(tchvec) ;
	    learn() ;
	 } ;

	 //--------------------------------------------------
	 // get sum of square error
      public:
	 Flt sumError() const { return output->sumError() ; } ;

	 //--------------------------------------------------
	 // get output
      public:
	 void getNetOutput(ValVector & outvec) const { 
	    output->getNetOutput(outvec) ;
	 } ;

	 //--------------------------------------------------
	 // scan
      public:
	 virtual void scan(StrForm & sform) {
	    BpNetBase::scan(sform) ;

	    context = layer(nameContext) ;
	    input = layer(nameInput) ;
	    inter1 = layer(nameInter1) ;
	    hidden = layer(nameHidden) ;
	    inter2 = layer(nameInter2) ;
	    output = layer(nameOutput) ;

	    con_int1 = link(context,inter1) ;
	    inp_int1 = link(input,inter1) ;
	    int1_hid = link(inter1,hidden) ;
	    hid_int2 = link(hidden,inter2) ;
	    int2_out = link(inter2,output) ;
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif


