// -*- Mode:C++ -*-
//Header:
//File: MentalWorld.h
//Title: mental World Model Network
//Author: NODA, Itsuki
//Date: 2001/07/01
//

//ModifyHistory:
// 2001/07/01: Start to create this file
//EndModifyHistory:

//Memo:
/*
 *  (in one time frame)
 *
 *              [sensor] <=> [expect] [action] 
 *                               |   
 *                           [inter2] [OLLout]
 *                               |   /
 *                           [hidden]
 *                               |   \
 *                           [inter1] [OLLin]
 *                          /    |   \
 *               [hidden-]  [action-] [sensor]
 *
 *  (in two time frames)
 *
 *                 [sensor_t+1] <=> [expect_t+1]
 *                                         \
 *                                         [inter2_t+1]
 *     [sensor_t] <=> [expect_t]           [hidden_t+1]
 *                           \             [inter1_t+1]
 *                          [inter2_t]    /        \
 *                               [hidden_t]        [action_t]
 *                               [inter1_t]
 *                                 /     \
 *                         [hidden_t-1][action_t-1]+[sensor_t]
 *
 *
 *  (in multi time frames)
 *
 *                        [expect_t+1]   
 *                                    \  
 *                   [ob] [expect_t]   [HIDDEN_t+1]
 *                                  \       /     \
 *                [ob] [expect_t-1]   [HIDDEN_t]    [action_t]
 *                                \       /    \       
 *              [ob]  [expect_t-2]  [HIDDEN_t-1]  [action_t-1]
 *                              \       /    \       
 *                              [HIDDEN_t-2]  [action_t-2]
 *                  
 */
//EndMemo:

#ifndef _Nnn_MentalWorld_h_
#define _Nnn_MentalWorld_h_
////////////////////////////////////////////////////////////////////////

#include "VBpNetBase.h"
#include "MwTrainSetBase.h"

namespace Nnn {
   using namespace Itk ;

   //======================================================================
   // class MentalWorld

   class MentalWorld : public WithDescriber {

	 //============================================================
	 // MentalWorld:: class Frame
      public:
	 class Frame : public VBpNetBase {

	       //--------------------------------------------------
	       // MentalWorld::Frame:: constants
	    public:
	       static char * const nameSensor = "sensor" ;
	       static char * const nameAction = "action" ;
	       static char * const nameInter1 = "inter1" ;
	       static char * const nameHidden = "hidden" ;
	       static char * const nameInter2 = "inter2" ;
	       static char * const nameExpect = "expect" ;
	       static char * const nameOLLinp = "ollinp" ;
	       static char * const nameOLLout = "ollout" ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: layers and links
	    public:
	       VBpLayer * sensor ;
	       VBpLayer * action ;
	       VBpLayer * inter1 ;
	       VBpLayer * hidden ;
	       VBpLayer * inter2 ;
	       VBpLayer * expect ;
	       
	       VBpLayer * ollinp ;
	       VBpLayer * ollout ;

	    public:
	       ValVector * senmask ;
	       ValVector * actmask ;

	    public:
	       VBpLLink * hid_int1 ;
	       VBpLLink * sen_int1 ;
	       VBpLLink * act_int1 ;
	       VBpLLink * int1_hid ;
	       VBpLLink * hid_int2 ;
	       VBpLLink * int2_exp ;

	       VBpLLink * hid_hid ;
	       VBpLLink * sen_hid ;
	       VBpLLink * act_hid ;
	       VBpLLink * hid_exp ;

	       VBpLLink * oll_hid ;
	       VBpLLink * hid_oll ;

	    public:
	       static const Val DfltOLLNoise = 1.0 ;
	       //static const Val DfltOLLFactor = 0.1 ; // ollout->erate

	       Bool ollp ;
	       Val ollnoise ;	/* scale of noise of oll */
	       //Val ollfactor ;  /* factor of OLL (in BP) */ // ollout->erate
	       

	       //--------------------------------------------------
	       // MentalWorld::Frame:: constructor
	    public:
	       Frame() { 
		  init() ; 
		  initLocal() ;
	       } ;

	       void initLocal() {
		  ollp = True ;
		  ollnoise = DfltOLLNoise ;
	       } ;
	       
	       //--------------------------------------------------
	       // MentalWorld::Frame:: setup
	    public:
	       void setup(const SString & nm,
			  const UInt & sensz,
			  const UInt & actsz,
			  const UInt & hidsz,
			  const UInt & intsz,
			  const Val & tempval = NNN_DFLT_TEMP,
			  const Val & epsval = NNN_DFLT_EPS,
			  const Val & alphaval = NNN_DFLT_ALPHA) {
		  init(nm,tempval,epsval,alphaval) ;

		  ollp = True ;
		  ollnoise = DfltOLLNoise ;

		  assignLayerLLink(sensz,actsz,hidsz,intsz) ;
	       } ;


	       //--------------------------------------------------
	       // MentalWorld::Frame:: alloc
	    public:
	       void assignLayerLLink(const UInt & sensz,
				     const UInt & actsz,
				     const UInt & hidsz,
				     const UInt & intsz) {

		  sensor = newLayer(nameSensor, sensz, VBpLayer::FT_linear);
		  action = newLayer(nameAction, actsz, VBpLayer::FT_linear);
		  hidden = newLayer(nameHidden, hidsz, VBpLayer::FT_linear);
		  expect = newLayer(nameExpect, sensz, VBpLayer::FT_linear);
		  inter1 = newLayer(nameInter1, intsz, VBpLayer::FT_sigmoid);
		  inter2 = newLayer(nameInter2, intsz, VBpLayer::FT_sigmoid);
		  
		  senmask = new ValVector(sensz) ;
		  actmask = new ValVector(actsz) ;

		  hid_int1 = newLLink(hidden, inter1) ;
		  sen_int1 = newLLink(sensor, inter1) ;
		  act_int1 = newLLink(action, inter1) ;
		  int1_hid = newLLink(inter1, hidden);
		  hid_int2 = newLLink(hidden, inter2) ;
		  int2_exp = newLLink(inter2, expect) ;

		  hid_hid = newLLink(hidden, hidden) ;
		  sen_hid = newLLink(sensor, hidden) ;
		  act_hid = newLLink(action, hidden) ;
		  hid_exp = newLLink(hidden, expect) ;

		  /* OLL part */

		  ollinp = newLayer(nameOLLinp, hidsz, VBpLayer::FT_linear);
		  ollout = newLayer(nameOLLout, hidsz, VBpLayer::FT_linear);

		  oll_hid = newLLink(ollinp, hidden) ;
		  hid_oll = newLLink(hidden, ollout) ;

	       } ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: scan
	    public:
	       virtual void scan(StrForm & sform) {
		  VBpNetBase::scan(sform) ;

		  sensor = layer(nameSensor) ;
		  action = layer(nameAction) ;
		  hidden = layer(nameHidden) ;
		  expect = layer(nameExpect) ;
		  inter1 = layer(nameInter1) ;
		  inter2 = layer(nameInter2) ;

		  senmask = new ValVector(sensor->n) ;
		  actmask = new ValVector(action->n) ;

		  hid_int1 = link(hidden, inter1) ;
		  sen_int1 = link(sensor, inter1) ;
		  act_int1 = link(action, inter1) ;
		  int1_hid = link(inter1, hidden);
		  hid_int2 = link(hidden, inter2) ;
		  int2_exp = link(inter2, expect) ;

		  hid_hid = link(hidden, hidden) ;
		  sen_hid = link(sensor, hidden) ;
		  act_hid = link(action, hidden) ;
		  hid_exp = link(hidden, expect) ;
		  
		  /* OLL part */

		  ollinp = layer(nameOLLinp) ;
		  ollout = layer(nameOLLout) ;
		  if(isNull(ollinp) || isNull(ollout)) {
		     ITK_WRN("run without OLL!!") ;
		     ollp = False ;
		  } else {
		     oll_hid = link(ollinp, hidden) ;
		     hid_oll = link(hidden, ollout) ;
		     if(isNull(oll_hid) || isNull(hid_oll)) {
			ITK_WRN("OLL part has not complete links.") ;
		     }			
		  }

	       } ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: config inter-frame link
	    public:
	       void configInterFrameLLink(Frame & preframe) {
		  return configInterFrameLLink(&preframe) ;
	       }  ;

	       void configInterFrameLLink(Frame * preframe) {
		  if(!isNull(preframe)) {
		     /* remove old from-layer's link */
		     hid_int1->from->removeToLink(hid_int1) ;
		     hid_hid->from->removeToLink(hid_hid) ;

		     /* add new links */
		     VBpLayer * newfrom = preframe->hidden ;

		     hid_int1->from = newfrom ;
		     newfrom->addToLink(hid_int1) ;

		     hid_hid->from = newfrom ;
		     newfrom->addToLink(hid_hid) ;

		  } 
		  else {	// if preframe is null, this is the initial
		     		// layer.
		     hid_int1->from->removeToLink(hid_int1) ;
		     hid_int1->to->removeFromLink(hid_int1) ;

		     hid_hid->from->removeToLink(hid_hid) ;
		     hid_hid->to->removeFromLink(hid_hid) ;
		  }
	       } ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: set outside values

	    public:
	       void setSensorAction(const ValVector & senvec,
				    const ValVector & actvec,
				    const ValVector & smask 
				    = *static_cast<ValVector*>(ITK_NULLPTR),
				    const ValVector & amask 
				    = *static_cast<ValVector*>(ITK_NULLPTR)) {
		  setSensor(senvec,smask) ;
		  setAction(actvec,amask) ;
	       } ;

	    public:
	       void setSensor(const ValVector & senvec,
			      const ValVector & smask 
			      = *static_cast<ValVector*>(ITK_NULLPTR),
			      const ValVector & auxmask
			      = *static_cast<ValVector*>(ITK_NULLPTR)) {
		  sensor->inp().copyFrom(senvec) ;
		  sensor->rinp().copyFrom(senvec) ;

		  if(!isNull(&smask)) {
		     senmask->copyFrom(smask) ;
		     /* masking to rinp() should be binary.
		      *  only mask is 0.0, the value is set 0.0.
		      */
		     //sensor->rinp().mask(smask) ;
		     sensor->rinp().bmask(smask) ;
		     if(!isNull(&auxmask)) {
			senmask->mask(auxmask) ;
			//sensor->rinp().mask(auxmask) ;
			sensor->rinp().bmask(auxmask) ;
		     }
		  }
		  else senmask->set(1.0) ;
	       } ;

	    public:
	       void setAction(const ValVector & actvec,
			      const ValVector & amask 
			      = *static_cast<ValVector*>(ITK_NULLPTR)) {
		  action->inp().copyFrom(actvec) ;
		  action->rinp().copyFrom(actvec) ;

		  if(!isNull(&amask)) {
		     actmask->copyFrom(amask) ;
		     action->rinp().mask(amask) ;
		  }
		  else actmask->set(1.0) ;
	       } ;

	    public:
	       void clearAction() {
		  action->inp().set(0.0) ;
		  action->rinp().set(0.0) ;
		  actmask->set(0.0) ;
	       } ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: prop foreward
	       //	suppose that 
	       //		previous hidden layer is set correctly.
	       //		sensor and action input are set inp()s.
	    public:
	       void propFore() {
		  sensor->propForeWithMask(sensor->inp(),*senmask,temp) ;
		  action->propForeWithMask(action->inp(),*actmask,temp) ;

		  if(ollp) {
		     ollinp->out().randomize(-ollnoise,ollnoise) ;
		  } else {
		     if(!isNull(ollinp)) {
			ollinp->out().set(0.0) ;
		     }
		  }

		  inter1->propFore(temp) ;
		  hidden->propFore(temp) ;
		  inter2->propFore(temp) ;
		  expect->propFore(temp) ;

		  if(ollp) {
		     ollout->propFore(temp) ;
		  }
	       } ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: prop back
	       //	suppose that
	       //		sensor input are set inp()s.
	    public:
	       void propBack() {
		  expect->propBackWithMask(sensor->inp(),*senmask,temp) ;

		  if(ollp) {
		     ollout->propBack(ollinp->out(),temp) ;
		  }

		  inter2->propBack(temp) ;
		  hidden->propBack(temp) ;
		  inter1->propBack(temp) ;
		  action->propBack(temp) ;
		  sensor->propBack(temp) ;
	       } ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: prop fore and back
	       //	used for the first frame (context frame)

	       void propForeFromHidden() {
		  inter2->propFore(temp) ;
		  expect->propFore(temp) ;
	       } ;

	       void propBackToHidden() {
		  expect->propBackWithMask(sensor->inp(),*senmask,temp) ;

		  inter2->propBack(temp) ;
		  hidden->propBack(temp) ;
	       } ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: learn
	    public:
	       Val learn() {
		  Val dsum = 0.0 ;

		  dsum += inter1->learn(eps,alpha) ;
		  dsum += hidden->learn(eps,alpha) ;
		  dsum += inter2->learn(eps,alpha) ;
		  dsum += expect->learn(eps,alpha) ;
		  
		  dsum += hid_int1->learn(eps,alpha) ;
		  dsum += sen_int1->learn(eps,alpha) ;
		  dsum += act_int1->learn(eps,alpha) ;
		  dsum += int1_hid->learn(eps,alpha) ;
		  dsum += hid_int2->learn(eps,alpha) ;
		  dsum += int2_exp->learn(eps,alpha) ;

		  dsum += hid_hid->learn(eps,alpha) ;
		  dsum += sen_hid->learn(eps,alpha) ;
		  dsum += act_hid->learn(eps,alpha) ;
		  dsum += hid_exp->learn(eps,alpha) ;

		  if(ollp) {
		     dsum += hid_oll->learn(eps,alpha) ;
		     dsum += oll_hid->learn(eps,alpha) ;
		  }

		  cntLearn++ ;
		  
		  return dsum ;
	       } ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: adapt action
	    public:
	       void adaptAction(Flt beta) {
		  action->adaptRInpByDelta(beta) ;
	       } ;
		  
	       //--------------------------------------------------
	       // MentalWorld::Frame:: adapt hidden (only for initial frame)
	    public:
	       void adaptHidden(Flt beta) {
		  hidden->adaptOutByDelta(beta) ;
	       } ;
		  
	       //--------------------------------------------------
	       // MentalWorld::Frame:: copyStatusFrom (used in shift)
	    public:
	       void copyStatusFrom(const Frame & frame) {
		  return copyStatusFrom(&frame) ;
	       } ;
	       void copyStatusFrom(const Frame * frame) {
		  hidden->out().copyFrom(frame->hidden->out()) ;

		  action->inp().copyFrom(frame->action->inp()) ;
		  action->rinp().copyFrom(frame->action->rinp()) ;
		  actmask->copyFrom(*frame->actmask) ;

		  sensor->inp().copyFrom(frame->sensor->inp()) ;
		  sensor->rinp().copyFrom(frame->sensor->rinp()) ;
		  senmask->copyFrom(*frame->senmask) ;
	       } ;

	       //--------------------------------------------------
	       // MentalWorld::Frame:: copyFrom
	    public:
	       void copyFrom(const Frame * original) {
		  setup(name,
			original->sensor->n,
			original->action->n,
			original->hidden->n,
			original->inter1->n,
			original->temp,original->eps,original->alpha) ;

		  sensor->copyFrom(*original->sensor,True,True) ;
		  action->copyFrom(*original->action,True,True) ;
		  inter1->copyFrom(*original->inter1,True,True) ;
		  hidden->copyFrom(*original->hidden,True,True) ;
		  inter2->copyFrom(*original->inter2,True,True) ;
		  expect->copyFrom(*original->expect,True,True) ;

		  hid_int1->copyFrom(*original->hid_int1,True,True) ;
		  act_int1->copyFrom(*original->act_int1,True,True) ;
		  sen_int1->copyFrom(*original->sen_int1,True,True) ;
		  int1_hid->copyFrom(*original->int1_hid,True,True) ;
		  hid_int2->copyFrom(*original->hid_int2,True,True) ;
		  int2_exp->copyFrom(*original->int2_exp,True,True) ;

		  hid_hid->copyFrom(*original->hid_hid,True,True) ;
		  sen_hid->copyFrom(*original->sen_hid,True,True) ;
		  act_hid->copyFrom(*original->act_hid,True,True) ;
		  hid_exp->copyFrom(*original->hid_exp,True,True) ;

		  /* OLL part */
		  
		  if(original->ollp) {
		     ollp = original->ollp ;
		     ollnoise = original->ollnoise ;
		     ollinp->copyFrom(*original->ollinp,True,True) ;
		     ollout->copyFrom(*original->ollout,True,True) ;
		     oll_hid->copyFrom(*original->oll_hid,True,True) ;
		     hid_oll->copyFrom(*original->hid_oll,True,True) ;
		  } else {
		     ollp = False ;
		  }

	       } ;
	 } ;
	 
	 //============================================================
	 // MentalWorld:: body part
	 
	 //--------------------------------------------------
	 // MentalWorld:: member
      public:
	 Frame original ;
	 Frame * frame ;
	 UInt n ;

	 Int worktime ;  /* time for frame[1] ??? 
			    time for frame[1].sensor & frame[2].action
			  */
	 
	 //--------------------------------------------------
	 // MentalWorld::Frame:: access
      public:
	 ITK_DEF_ACCESS(UInt & cntLearn(),{return original.cntLearn;});
	 ITK_DEF_ACCESS(UInt & cntTrial(),{return original.cntTrial;});

	 //--------------------------------------------------
	 // MentalWorld:: constructor
      public:
	 MentalWorld() {
	    init() ;
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: init
      public:
	 void init() {
	    n = 0 ;
	    frame = ITK_NULLPTR ;
	    worktime = -1 ;
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: load
      public:
	 virtual void load(const SubString & file,Bool expandp = False) {
	    original.load(file) ;
	    if(expandp) expand() ;
	 } ;
	 virtual void load(istream & istr, Bool expandp = False) {
	    original.load(istr) ;
	    if(expandp) expand() ;
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: save
      public:
	 virtual void save(const SubString& file, 
			   SaveLevel slevel = SL_NetState) {
	    original.save(file,slevel) ;
	 } ;
	 virtual void save(ostream & ostr, 
			   SaveLevel slevel = SL_NetState,
			   const Bool & newlinep = True,
			   const SubString & indent = SubString("")) {
	    original.save(ostr,slevel,newlinep,indent) ;
	 } ;
	 
	 //--------------------------------------------------
	 // MentalWorld:: dump
      public:
	 virtual void dump(const SubString& file) {
	    ofstream ostr(file) ;
	    dump(ostr) ;
	 } ;
	 virtual void dump(ostream & ostr) {
	    for(UInt i = 0 ; i < n ; i++) {
	       ostr << "<<<<<<<<<<< " << i << ">>>>>>>>>>>>>>" << endl ;
	       frame[i].save(ostr,SL_AllState,True) ;
	    }
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: expand
      public:
	 virtual void expand(UInt N = 0) {
	    if(N > n) {
	       n = N ;
	       if(!isNull(frame)) delete frame ;
	       frame = new Frame[n] ;
	    }
	    for(UInt i = 0 ; i < n ; i++) {
	       frame[i].copyFrom(&original) ;
	    }
	    for(UInt i = 0 ; i < n ; i++) {
	       if(i == 0) {
		  frame[i].configInterFrameLLink(ITK_NULLPTR) ;
	       } else {
		  frame[i].configInterFrameLLink(frame[i-1]) ;
	       }
	    }
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: shift
      public:
	 virtual void shift(UInt m = 1, UInt N = -1, Bool countUp = True) {
	    if(N < 0) N = n ;
	    for(UInt i = 0 ; i < N - m ; i++) {
	       frame[i].copyStatusFrom(frame[i+m]) ;
	    }
	    if(countUp) worktime += m ;
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: propFore 
	 // 	prop fore the first m frame
      public:
	 void propFore(UInt m, UInt base = 0) {
	    if(m > n || base > m) {
	       ITK_ERR("m (=" << m << ") should be less than n (=" 
		       << n << ") and more than base (=" << base << ".") ;
	       throw 0 ;
	    }
	    for(UInt i = base ; i < m ; i++) {
	       if(i == base) {
		  frame[i].propForeFromHidden() ;
	       } else {
		  frame[i].propFore() ;
	       }
	    } ;
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: propBack
      public:
	 void propBack(UInt m, UInt base = 0) {
	    if(m > n || base > m) {
	       ITK_ERR("m (=" << m << ") should be less than n (=" 
		       << n << ") and more than base (=" << base << ".") ;
	       throw 0 ;
	    }
	    for(UInt i = m ; i > base ; i--) {
	       if(i-1 == base) {
		  frame[i-1].propBackToHidden() ;
	       } else {
		  frame[i-1].propBack() ;
	       }
	    } ;
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: adapt
      public:
	 void adapt(UInt m, Flt beta, UInt base = 0) {
	    if(m > n || base > m) {
	       ITK_ERR("m (=" << m << ") should be less than n (=" 
		       << n << ") and more than base (=" << base << ".") ;
	       throw 0 ;
	    }
	    for(UInt i = base ; i < m ; i++) {
	       if(i == base) {
		  frame[i].adaptHidden(beta) ;  
	       } else {
		  frame[i].adaptAction(beta) ;
	       }
	    }
	 }

	 //--------------------------------------------------
	 // MentalWorld:: propAdapt
	 //		propFore, propBack and adapt
	 //		suppose that all sensor and action are set
      public:
	 void propAdapt(UInt n, UInt adaptN, Val beta) {
	    for(UInt i = 0 ; i < adaptN ; i++) {
	       propFore(n) ;
	       propBack(n) ;
	       adapt(n,beta) ;
	    }
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: learn
      public:
	 Val learn(UInt m, UInt base = 0) {
	    Val dsum = 0.0 ;
	    
	    for(UInt i = base + 1 ; i < m ; i++) {
	       dsum += frame[i].learn() ;
	    } ;
	    cntLearn()++ ;
	    return dsum ;
	 } ;

	 //------------------------------------------------------------
	 // MentalWorld:: Run 2 Frame (old)

	 //--------------------------------------------------
	 // MentalWorld:: run 2 frame (old style)
	 //	just run 0-2 processing
      public:
	 void run2frame(UInt adaptN,
			const ValVector & curact,
			const ValVector & cursen,
			const ValVector & cactmask 
			= *static_cast<ValVector*>(ITK_NULLPTR),
			const ValVector & csenmask 
			= *static_cast<ValVector*>(ITK_NULLPTR)) {
	    shift(1,3) ;
	    frame[1].setSensor(cursen,csenmask) ;
	    frame[2].setAction(curact,cactmask) ;

	    for(UInt i = 0 ; i < adaptN ; i++) {
	       /* activate and adapt current time frame */
	       propFore(2) ;		
	       propBack(2) ;
	       adapt(2,0.1) ;		////**** xxxxxx BETA value ****
	    }
	    /* activate to the next time frame */
	    propFore(3) ;

	    cntTrial()++ ;
	 }
	 //--------------------------------------------------
	 // MentalWorld:: run & learn 2 frame (old style)
	 //	just run 0-2 processing
      public:
	 Val runLearn2frame(UInt adaptN,
			     const ValVector & curact,
			     const ValVector & cursen,
			     const ValVector & nxtsen,
			     const ValVector & cactmask 
			     = *static_cast<ValVector*>(ITK_NULLPTR),
			     const ValVector & csenmask 
			     = *static_cast<ValVector*>(ITK_NULLPTR),
			     const ValVector & nsenmask 
			     = *static_cast<ValVector*>(ITK_NULLPTR)) {
	    run2frame(adaptN,curact,cursen,cactmask,csenmask) ;

	    frame[2].setSensor(nxtsen,nsenmask) ;
	    propBack(3) ;
	    return learn(3) ;
	 }

	 //------------------------------------------------------------
	 // MentalWorld:: Run 2 Frame with MwTrainSetBase

	 //--------------------------------------------------
	 // MentalWorld:: run 2 frame with MwTrainSetBase
	 //	just run 0-2 processing
      public:
	 void run2frame(MwTrainSetBase & tset,
		       UInt adaptN = 5, Val beta = 0.1) {
	    shift(1,3) ;
	    
	    MwTrainSetBase::Frame & cset = tset.getFrame(worktime) ;
	    frame[1].setSensor(cset.sensor(),cset.senMask(),tset.inpMask()) ;
	    frame[2].setAction(cset.action(),cset.actMask()) ;

	    Bool orig_ollp = switchOLL(False) ; // during adapt, OLL off
	    { 
	       propAdapt(2,adaptN,beta) ;
	    }
	    switchOLL(orig_ollp) ;

	    propFore(3) ;

	    cntTrial()++ ;
	 }

	 //--------------------------------------------------
	 // MentalWorld:: access to teacher (trainset) value 
      public:
	 ValVector & sensorTch(MwTrainSetBase & tset, Int framenum) {
	    return tset.getFrame(worktime + framenum - 1).sensor() ;
	 } 
	 ValVector & actionTch(MwTrainSetBase & tset, Int framenum) {
	    return tset.getFrame(worktime + framenum - 1).action() ;
	 } 

	 //--------------------------------------------------
	 // MentalWorld:: run & learn 2 frame with MwTrainSetBase
	 //	just run 0-2 processing
      public:
	 Val runLearn2frame(MwTrainSetBase & tset,
			    UInt adaptN = 5, Val beta = 0.1) {
	    run2frame(tset,adaptN,beta) ;
	    
	    MwTrainSetBase::Frame & nset = tset.getFrame(worktime+1) ;
	    frame[2].setSensor(nset.sensor(),nset.senMask()) ;
	    propBack(3) ;
	    return learn(3) ;
	 } ;

	 //------------------------------------------------------------
	 // MentalWorld:: Run N Frame with MwTrainSetBase

	 //--------------------------------------------------
	 // MentalWorld:: run N frame with MwTrainSetBase
	 //	run 0-N processing
      public:
	 void runNframe(MwTrainSetBase & tset, UInt n,
			UInt adaptN = 5, Val beta = 0.1) {
	    shift(1,n+1) ;

	    MwTrainSetBase::Frame & cset = tset.getFrame(worktime) ;
	    frame[n-1].setSensor(cset.sensor(),cset.senMask(),tset.inpMask()) ;
	    frame[n].setAction(cset.action(),cset.actMask()) ;

	    /* clearn up sensor for frame[n] */
	    frame[n].sensor->inp().set(0.0) ;
	    frame[n].sensor->rinp().set(0.0) ;
	    
	    Bool orig_ollp = switchOLL(False) ; {
	       for(UInt i = 0 ; i < adaptN ; i++) {
		  propFore(n) ;
		  propBack(n) ;
		  adapt(n,beta) ;
	       }
	    }
	    switchOLL(orig_ollp) ;

	    propFore(n+1) ;
	    cntTrial()++ ;
	 }

	 //--------------------------------------------------
	 // MentalWorld:: run & learn N frame with MwTrainSetBase
	 //	just run 0-N processing
      public:
	 Val runLearnNframe(MwTrainSetBase & tset, UInt n,
			    UInt adaptN = 5, Val beta = 0.1) {
	    runNframe(tset,n,adaptN,beta) ;
	    
	    MwTrainSetBase::Frame & nset = tset.getFrame(worktime+1) ;
	    frame[n].setSensor(nset.sensor(),nset.senMask()) ;
	    propBack(n+1) ;
	    return learn(n+1) ;
	 } ;

	 //------------------------------------------------------------
	 // MentalWorld:: Planning

	 //--------------------------------------------------
	 // MentalWorld:: simple planning
	 //		just to reach goal
      public:
	 void simplePlanning(ValVector & cSensor, ValVector & cMask,
			     ValVector & gSensor, ValVector & gMask,
			     Int duration, UInt adaptN, Val beta) {
	    ValVector wAction(original.action->n) ;
	    ValVector wSensor(original.sensor->n) ;
	    ValVector actOffMask(original.action->n) ;
	    ValVector senOffMask(original.sensor->n) ;

	    wAction.set(0.0) ;
	    wSensor.set(0.0) ;
	    actOffMask.set(0.0) ;
	    senOffMask.set(0.0) ;

	    // set initial state
	    frame[0].setSensor(cSensor,senOffMask) ;
	    frame[1].setSensor(cSensor,cMask) ;
	    frame[1].setAction(wAction,actOffMask) ;

	    // set intermediate frames
	    for(Int i = 2 ; i < duration + 2 ; i++) {
	       frame[i].setAction(wAction,actOffMask) ;
	       frame[i].setSensor(wSensor,senOffMask) ;
	    }
	    // set goal frames
	    frame[duration + 1].setSensor(gSensor,gMask) ;

	    // OLL switch off
	    Bool orig_ollp = switchOLL(False) ; { // during planning
	       //make plan!!
	       for(UInt i = 0 ; i < adaptN ; i++) {
		  propFore(duration+2) ;
		  propBack(duration+2) ;
		  adapt(duration+2, beta) ;
	       }
	    }
	    switchOLL(orig_ollp) ;
	 } ;

	 //--------------------------------------------------
	 // MentalWorld:: constrainted planning
      public:
	 void constraintedPlanning(MwTrainSetBase & tset,
				   Int duration, UInt adaptN, Val beta) {

	    Bool orig_ollp = switchOLL(False) ; {
	       for(UInt i = 0 ; i < adaptN ; i++) {
		  for(Int j = 0 ; j < duration + 2 ; j++) {
		     Int wtime = worktime + j - 1 ;
		     MwTrainSetBase::Frame cset =
			tset.getConstraint(wtime,(i==0),
					   frame[j].sensor->out(),
					   frame[j+1].action->rinp()) ;
		     frame[j].setSensor(cset.sensor(),cset.senMask(),
					tset.inpMask()) ;
		     frame[j+1].setAction(cset.action(),cset.actMask()) ;
		  }
		  propFore(duration+2) ;
		  propBack(duration+2) ;
		  adapt(duration+2,beta) ;
	       }
	    } ; switchOLL(orig_ollp) ;
	 } ;
	       

	 //--------------------------------------------------
	 // MentalWorld:: control OLL
      public:
	 Bool switchOLL(Bool onoff) {
	    Bool r = original.ollp ;
	    
	    original.ollp = onoff ;
	    for(UInt i = 0 ; i < n ; i++) {
	       frame[i].ollp = onoff ;
	    }
	    return r ;
	 } ;

   } ;
} ;
////////////////////////////////////////////////////////////////////////
#endif

