// -*- Mode:C++ -*-
//Header:
//File: MwTrainSetBase.h
//Title: training set for mental World Model Network
//Author: NODA, Itsuki
//Date: 2001/07/09
//

//ModifyHistory:
// 2001/07/09: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_MwTrainSetBase_h_
#define _Nnn_MwTrainSetBase_h_
////////////////////////////////////////////////////////////////////////

namespace Nnn {
   using namespace Itk ;

   //======================================================================
   // class MwTrainSet
   //	Usage:
   //		Following virtual functions should be defined in a derived
   //		class:
   //			virtual const char* name() const ;
   //			virtual Int sensorN() const ;
   //			virtual Int actionN() const ;
   //			virtual void cycle(ValVector & sensor, 
   //					   ValVector & senMask,
   //					   ValVector & action, 
   //					   ValVector & actMask) ;


   class MwTrainSetBase : public WithDescriber {

	 //------------------------------------------------------------
	 // MwTrainSetBase:: virtual constant params
      public:
	 virtual const char* name() const = 0 ;
	 virtual Int sensorN() const = 0 ;
	 virtual Int actionN() const = 0 ;

	 const static Int DfltFrameBufSize = 256 ;

	 //============================================================
	 // MwTrainSetBase:: class Frame
      public:
	 class Frame : public WithDescriber {

	       //--------------------------------------------------
	       // MwTrainSetBase::Frame:: member
	    public:
	       ValVector * sensor_ ;
	       ValVector * action_ ;
	       ValVector * senMask_ ;
	       ValVector * actMask_ ;

	       //--------------------------------------------------
	       // MwTrainSetBase::Frame:: access
	    public:
	       ITK_DEF_ACCESS(ValVector & sensor(), { return *sensor_ ; }) ;
	       ITK_DEF_ACCESS(ValVector & action(), { return *action_ ; }) ;
	       ITK_DEF_ACCESS(ValVector & senMask(), { return *senMask_ ; }) ;
	       ITK_DEF_ACCESS(ValVector & actMask(), { return *actMask_ ; }) ;

	       //--------------------------------------------------
	       // MwTrainSetBase::Frame:: constructor
	    public:
	       Frame() { init() ; } ;
	       Frame(UInt sensorN, UInt actionN) { init(sensorN, actionN) ; } ;
	       
	       //--------------------------------------------------
	       // MwTrainSetBase::Frame:: init
	    public:
	       void init(UInt sensorN = 0, UInt actionN = 0) {
		  if(sensorN == 0) {
		     sensor_ = senMask_ = ITK_NULLPTR ;
		  } else {
		     sensor_ = new ValVector(sensorN) ;
		     senMask_ = new ValVector(sensorN) ;
		  } 

		  if(actionN == 0) {
		     action_ = actMask_ = ITK_NULLPTR ;
		  } else {
		     action_ = new ValVector(actionN) ;
		     actMask_ = new ValVector(actionN) ;
		  }
	       } ;
	 } ;

	 //------------------------------------------------------------
	 // MwTrainSetBase:: member
      public:
	 Frame * frame_ ;
	 Int timehead ;
	 Int frameBufSize ;

	 ValVector * inpMask_ ; /* masking sensor for input frames
				   to separate recognition part */

	 Frame * cframe_ ;	/* not used now*/

	 Frame workframe_ ;

	 //------------------------------------------------------------
	 // MwTrainSetBase:: access /* not used now */
      public:
	 void setCFrame(Int time) { cframe_ = &frame(time) ;} ;
	 ITK_DEF_ACCESS(Frame & cFrame(), { return *cframe_ ; }) ;
	 ITK_DEF_ACCESS(ValVector & cSensor(), { return cFrame().sensor();}) ;
	 ITK_DEF_ACCESS(ValVector & cAction(), { return cFrame().action();}) ;
	 ITK_DEF_ACCESS(ValVector & cSenMask(), { return cFrame().senMask();});
	 ITK_DEF_ACCESS(ValVector & cActMask(), { return cFrame().actMask();});

	 ITK_DEF_ACCESS(Frame & workframe(), { return workframe_ ; }) ;

	 //------------------------------------------------------------
	 // MwTrainSetBase:: access /* used now */
      public:
	 ITK_DEF_ACCESS(ValVector & inpMask(), { return *inpMask_ ; }) ;

	 //------------------------------------------------------------
	 // MwTrainSetBase:: construct
      public:
	 MwTrainSetBase() { } ;
	 
	 //------------------------------------------------------------
	 // MwTrainSetBase:: construct
      public:
	 virtual void init(Int bufsize = DfltFrameBufSize) { 
	    nop() ;
	    frameBufSize = bufsize ;
	    frame_ = new Frame[frameBufSize] ;
	    for(Int i = 0 ; i < frameBufSize ; i++) {
	       frame(i).init(sensorN(),actionN()) ;
	    }
	    timehead = 0 ;

	    inpMask_ = new ValVector(sensorN()) ;
	    inpMask_->set(1.0) ;	/* Normally, 
					   inpMask() = [1,1,....] */
	 } ;
	 
	 //------------------------------------------------------------
	 // MwTrainSetBase:: cycle
	 //	should be defined in derived classes
      public:
	 virtual void cycle(ValVector & sensor, ValVector & senMask,
			    ValVector & action, ValVector & actMask) = 0 ;

	 //------------------------------------------------------------
	 // MwTrainSetBase:: access (do not use)
      public:
	 virtual Frame & frame(Int time) {
	    return frame_[time % frameBufSize] ;
	 } ;

	 //------------------------------------------------------------
	 // MwTrainSetBase:: get frame (use this to get frame)
      public:
	 virtual Frame & getFrame(Int time) {
	    if(time < 0 || time <= (timehead - frameBufSize)) {
	       ITK_ERR("Illegal Frame Access: too small time") ;
	       ITK_DBG(time) ;
	       ITK_DBG(timehead) ;
	       ITK_DBG(frameBufSize) ;
	       throw 0 ;
	    } ;
	    while(timehead <= time) {
	       Frame & f = frame(timehead) ;
	       cycle(f.sensor(),f.senMask(),f.action(),f.actMask()) ;
	       timehead++ ;
	    }
	    return frame(time) ;
	 } ;

	 //------------------------------------------------------------
	 // MwTrainSetBase:: get contraint (for planning)
      public:
	 virtual Frame & getConstraint(Int time, Bool initp,
				       const ValVector & sensor, 
				       const ValVector & action) {
	    if(time < timehead) return frame(time) ;
	    else {
	       Frame & f = workframe() ;
	       setConstraint(f, time, initp, sensor, action) ;
	       return f ;
	    }
	 } ;

	 virtual void setConstraint(Frame & f, Int time, Bool initp,
				       const ValVector & sensor, 
				       const ValVector & action) {
	    ITK_TRC0() ;
	    ITK_WRN("This should be defined by user.") ;

	    f.sensor().set(0.0) ;	/* !!! */
	    f.senMask().set(1.0) ;	/* !!! */
	    
	    if(initp) {
	       f.action().randomize(-1.0,1.0) ; /* !!! */
	       f.actMask().set(1.0) ; 		/* !!! */
	    } else {
	       f.action().copyFrom(action) ;    /* !!! */
	       f.actMask().set(1.0) ;		/* !!! */
	    }
	 } ;

	 //------------------------------------------------------------
	 // MwTrainSetBase:: describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#MwTrainSet[" << name() << "|"
		 << "timehead=" << timehead 
		 << "/" << frameBufSize
		 << "]" ;
	    if(detailp) {
	       ostr << "\t" << "sensorN=" << sensorN() << endl 
		    << "\t" << "actionN=" << actionN() << endl ;
	    } 
	 } ;
   } ;
} ;


	 
////////////////////////////////////////////////////////////////////////
#endif

