// -*- Mode:C++ -*-
//Header:
//File: VBpLayer.cc
//Author: NODA, Itsuki
//Date: 2000/01/12
//

//ModifyHistory:
// 2000/01/12: Start to create this file
//EndModifyHistory:

#include "VBpLayer.h"

namespace Nnn {

   //======================================================================
   // Execution

   //----------------------------------------------------------------------
   // forward processing

   //--------------------------------------------------
   // set net input (set Input Vector to input layer)

   void VBpLayer::setNetInput(const ValVector & inpvec) {
      setOut(inpvec) ;
   } 

   //--------------------------------------------------
   // set net input with mask
   //    out = mask * inpvec + (1-mask) * rinp()

   void VBpLayer::setNetInputWithMask(const ValVector & inpvec, 
				      const ValVector & mask) {
      VIterator ov = out().begin() ;
      C_VIterator iv = inpvec.begin() ;
      C_VIterator mv = mask.begin() ;
      C_VIterator rv = rinp().begin() ;
      
      for( ; iv < inpvec.end() ; ov++, iv++, mv++, rv++) {
	 (*ov) = (*mv) * (*iv) + (1.0 - *mv) * (*rv) ;
      }
   } 

   //--------------------------------------------------
   // get net output

   void VBpLayer::getNetOutput(ValVector & outvec) {
      getOut(outvec) ;
   }

   //--------------------------------------------------
   // propFore

   void VBpLayer::propFore(const ValVector & inpvec, const Val & temp) {
      return setNetInput(inpvec) ;
   } ;

   void VBpLayer::propFore(const Val & temp) {
      switch(ftype) {
	 case FT_linear : return propForeLinear(temp) ;
	 case FT_sigmoid : return propForeSigmoid(temp) ;
	 case FT_bigmoid : return propForeBigmoid(temp) ;
	 default : throw 0 ;
      }
   } ;

   void VBpLayer::propForeLinear(const Val & temp) {
      collectInput(True) ;
      activateByLinear(temp) ;
   } ;

   void VBpLayer::propForeSigmoid(const Val & temp) {
      collectInput(True) ;
      activateBySigmoid(temp) ;
   } ;

   void VBpLayer::propForeBigmoid(const Val & temp) {
      collectInput(True) ;
      activateByBigmoid(temp) ;
   } ;

   //--------------------------------------------------
   // propFore with mask

   void VBpLayer::propForeWithMask(const ValVector & inpvec, 
				   const ValVector & mask,
				   const Val & temp) {
      return setNetInputWithMask(inpvec,mask) ;
   } ;

   

   //--------------------------------------------------
   // clear input to same value

   void VBpLayer::clearInput(const Val & value) {
      //for(Iterator u = begin() ; u < end() ; u++) (*u).inp = value ; 
      /*
        for(VIterator iv = inp().begin() ; iv < inp().end() ; iv++) 
      	 *iv = value ;
      */
      inp().setAll(value) ;
   } 

   //--------------------------------------------------
   // set input vector by threshold

   void VBpLayer::clearInputByThreshold() {
      //for(Iterator u = begin() ; u < end() ; u++) (*u).inp = (*u).thr ;
      VIterator hv = thr().begin() ;
      VIterator iv = inp().begin() ;
      for(; iv < inp().end() ; iv++, hv++) *iv = *hv ;
   } 

   //--------------------------------------------------
   // collect input

   void VBpLayer::collectInput(const Bool clearp) {
      if(clearp) clearInputByThreshold() ;

      for(LLinkList::Iterator lnk = from.begin() ; lnk < from.end() ; lnk++) 
	 (*lnk)->propFore() ;
   }
      
   //--------------------------------------------------
   // activate

   void VBpLayer::activateByLinear(const Val & temp) {
      /*
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).activateByLinear(temp) ;
	 */
      VIterator ov = out().begin() ;
      VIterator iv = inp().begin() ;
      for(; iv < inp().end() ; iv++, ov++) *ov = *iv / temp ;
   }

   void VBpLayer::activateBySigmoid(const Val & temp) {
      /*
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).activateBySigmoid(temp) ;
	 */
      VIterator ov = out().begin() ;
      VIterator iv = inp().begin() ;
      for(; iv < inp().end() ; iv++, ov++) *ov = sigmoid(*iv,temp) ;
   }

   void VBpLayer::activateByBigmoid(const Val & temp) {
      /*
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).activateByBigmoid(temp) ;
	 */
      VIterator ov = out().begin() ;
      VIterator iv = inp().begin() ;
      for(; iv < inp().end() ; iv++, ov++) *ov = bigmoid(*iv,temp) ;
   }
   
   //----------------------------------------------------------------------
   // Backward processing

   //--------------------------------------------------
   // calc net output error (for output layer)

   void VBpLayer::propBack(const ValVector & tchvec, const Val & temp) {
      switch(ftype) {
	 case FT_linear : return propBackLinear(tchvec,temp) ;
	 case FT_sigmoid : return propBackSigmoid(tchvec,temp) ;
	 case FT_bigmoid : return propBackBigmoid(tchvec,temp) ;
	 default : throw 0 ;
      }
   }

   void VBpLayer::propBackWithMask(const ValVector & tchvec, 
				   const ValVector & mask,
				   const Val & temp) {
      calcNetErrorWithMask(tchvec,mask) ;
      switch(ftype) {
	 case FT_linear : return calcDeltaByLinear(temp) ;
	 case FT_sigmoid : return calcDeltaBySigmoid(temp) ;
	 case FT_bigmoid : return calcDeltaByBigmoid(temp) ;
	 default : throw 0 ;
      }
   }

   void VBpLayer::propBackLinear(const ValVector & tchvec,
				const Val & temp) {
      calcNetError(tchvec) ;
      calcDeltaByLinear(temp) ;
   }

   void VBpLayer::propBackSigmoid(const ValVector & tchvec,
				const Val & temp) {
      calcNetError(tchvec) ;
      calcDeltaBySigmoid(temp) ;
   } ;

   void VBpLayer::propBackBigmoid(const ValVector & tchvec,
				 const Val & temp) {
      calcNetError(tchvec) ;
      calcDeltaByBigmoid(temp) ;
   } ;

   //--------------------------------------------------
   // prop backward (for hidden layer)

   void VBpLayer::propBack(const Val & temp) {
      collectDelta() ;
      switch(ftype) {
	 case FT_linear : propBackLinear(temp) ; break ;
	 case FT_sigmoid : propBackSigmoid(temp) ; break ;
	 case FT_bigmoid : propBackBigmoid(temp) ; break ;
	 default : throw 0 ;
      }
      if(satrate > 0.0) addSaturateStress() ;
   }

   void VBpLayer::propBackLinear(const Val & temp) {
      calcDeltaByLinear(temp) ;
   } ;

   void VBpLayer::propBackSigmoid(const Val & temp) {
      calcDeltaBySigmoid(temp) ;
   } ;

   void VBpLayer::propBackBigmoid(const Val & temp) {
      calcDeltaByBigmoid(temp) ;
   } ;

   //--------------------------------------------------
   // set net output error

   void VBpLayer::calcNetError(const ValVector & tchvec) {
      C_VIterator tv = tchvec.begin() ;
      VIterator ev = err().begin() ;
      VIterator ov = out().begin() ;
      for(; ov < out().end() ; ov++, tv++, ev++) *ev = erate * (*tv - *ov) ;
   } 

   //--------------------------------------------------
   // set net output error

   void VBpLayer::calcNetErrorWithMask(const ValVector & tchvec,
				       const ValVector & mask) {
      C_VIterator tv = tchvec.begin() ;
      C_VIterator mv = mask.begin() ;
      VIterator ev = err().begin() ;
      VIterator ov = out().begin() ;
      for(; ov < out().end() ; ov++, mv++, tv++, ev++) 
	 *ev = erate * (*tv - *ov) * (*mv);
   } 

   //--------------------------------------------------
   // clear Error
   void VBpLayer::clearError() {
      for(VIterator ev = err().begin() ; ev < err().end() ; ev++) 
	 *ev = 0.0 ;
   } 

   //--------------------------------------------------
   // sum of error

   Val VBpLayer::sumError() const {
      Val sqerr = 0.0 ;
      for(C_VIterator ev = err().begin() ; ev < err().end() ; ev++) 
	 sqerr += sq(*ev) ;
      return sqerr ;
   } 

   //--------------------------------------------------
   // collect delta-in

   void VBpLayer::collectDelta() {
      clearError() ;
      for(LLinkList::Iterator lnk = to.begin() ; lnk < to.end() ; lnk++)
	 (*lnk)->propBack() ;
   }
      
   //--------------------------------------------------
   // calc delta

   void VBpLayer::calcDeltaByLinear(const Val & temp) {
      /*
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).calcDeltaByLinear(temp) ;
	 */
      VIterator dv = dlt().begin() ;
      VIterator ev = err().begin() ;
      for(; ev < err().end() ; ev++, dv++)
	 *dv = *ev / temp ;
   }

   void VBpLayer::calcDeltaBySigmoid(const Val & temp) {
      /*
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).calcDeltaBySigmoid(temp) ;
	 */
      VIterator dv = dlt().begin() ;
      VIterator iv = inp().begin() ;
      VIterator ov = out().begin() ;
      VIterator ev = err().begin() ;
      for(; ev < err().end() ; ev++, dv++, iv++, ov++)
	 *dv = *ev * d_sigmoid(*iv,*ov,temp) ;
   } 

   void VBpLayer::calcDeltaByBigmoid(const Val & temp) {
      /*
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).calcDeltaByBigmoid(temp) ;
	 */
      VIterator dv = dlt().begin() ;
      VIterator iv = inp().begin() ;
      VIterator ov = out().begin() ;
      VIterator ev = err().begin() ;
      for(; ev < err().end() ; ev++, dv++, iv++, ov++)
	 *dv = *ev * d_bigmoid(*iv,*ov,temp) ;
   } 

   //--------------------------------------------------
   // add SaturateStress

   void VBpLayer::addSaturateStress() {
      VIterator dv = dlt().begin() ;
      VIterator iv = inp().begin() ;
      for(; dv < dlt().end() ; dv++, iv++) {
	 if(*iv > satval) {
	    *dv -= (satrate * (*iv - satval)) ;
	 } else if (*iv < -satval) {
	    *dv -= (satrate * (*iv + satval)) ;
	 } else {
	 }
      }
   }

   //--------------------------------------------------
   // adapt Reveised Input by Delta (for adaptation by BP) (for input layer)
   
   void VBpLayer::adaptRInpByDelta(const Val & alpha) {
      VIterator rv = rinp().begin() ;
      C_VIterator ov = out().begin() ;
      C_VIterator dv = dlt().begin() ;
      for( ; rv < rinp().end() ; rv++, ov++, dv++) {
	 (*rv) = (*ov + alpha * (*dv)) ;
      }
   }

   void VBpLayer::adaptOutByDelta(const Val & alpha) {
      VIterator ov = out().begin() ;
      C_VIterator dv = dlt().begin() ;
      for( ; ov < out().end() ; ov++, dv++) {
	 (*ov) = (*ov + alpha * (*dv)) ;
      }
   }
      

   //----------------------------------------------------------------------
   // learn

   Val VBpLayer::learn(const Val & eps, const Val & alpha) {
      Val dsum = 0 ;

      Val e = eps * lrate ;

      VIterator dhv = dh().begin() ;
      VIterator dv = dlt().begin() ;
      VIterator hv = thr().begin() ;

      for(; dhv < dh().end() ; dhv++, dv++, hv++) {
	 (*dhv) = alpha * (*dhv) + e * (*dv) ;
	 (*hv) += (*dhv) ;
	 dsum += sq(*dhv) ;
	 // for debug to avoid overflow
	 //if(Abs(*dhv) > TooBigChange) NNN_XPT(0,"too big dw",this) ;
      }
      return dsum ;
   }

   //----------------------------------------------------------------------
   // adjust output by error value

   void VBpLayer::adjustOutputByError(const Val & eps) {
      VIterator ov = out().begin() ;
      VIterator ev = err().begin() ;
      for(; ov < out().end() ; ov++, ev++) 
	 (*ov) += (eps * (*ev)) ;
   }

   //======================================================================
   // Preparation

   //----------------------------------------------------------------------
   // add from/to link information

   Bool VBpLayer::addFromLink(VBpLLink * lnk) { 
      if(isNull(from.findPtr(lnk))) {
	 from.push(lnk) ; 
	 return True ;
      } else {
	 return False ;
      }
   } ;      

   Bool VBpLayer::addToLink(VBpLLink * lnk) {
      if(isNull(to.findPtr(lnk))) {
	 to.push(lnk) ; 
	 return True ;
      } else {
	 return False ;
      } 
   } ;

   Bool VBpLayer::removeToLink(VBpLLink * lnk) {
      Int c = to.removeAll(lnk) ;
      return (c>0) ;
   } ;

   Bool VBpLayer::removeFromLink(VBpLLink * lnk) {
      Int c = from.removeAll(lnk) ;
      return (c>0) ;
   } ;

   //----------------------------------------------------------------------
   // randomize threshold

   void VBpLayer::randomizeThreshold(const Val & minval, const Val & maxval) {
      randomizeByTime() ;
      for(VIterator hv = thr().begin() ; hv < thr().end() ; hv++) {
	 (*hv) = fltRand(minval,maxval) ;
      }
   }

   //----------------------------------------------------------------------
   // access to each vector

   void VBpLayer::setInp(const ValVector & inpvec) { 
      inp().copyFrom(inpvec) ; 
   } ;
   void VBpLayer::getInp(ValVector & inpvec) const { 
      inp().copyTo(inpvec) ; 
   } ;

   void VBpLayer::setOut(const ValVector & outvec) { 
      out().copyFrom(outvec) ; 
   } ;
   void VBpLayer::getOut(ValVector & outvec) const { 
      out().copyTo(outvec) ; 
   } ;

   void VBpLayer::setThr(const ValVector & thrvec) { 
      thr().copyFrom(thrvec) ; 
   } ;
   void VBpLayer::getThr(ValVector & thrvec) const { 
      thr().copyTo(thrvec) ; 
   } ;

   void VBpLayer::setErr(const ValVector & errvec) { 
      err().copyFrom(errvec) ; 
   } ;
   void VBpLayer::getErr(ValVector & errvec) const { 
      err().copyTo(errvec) ; 
   } ;

   void VBpLayer::setDlt(const ValVector & dltvec) { 
      dlt().copyFrom(dltvec) ; 
   } ;
   void VBpLayer::getDlt(ValVector & dltvec) const { 
      dlt().copyTo(dltvec) ; 
   } ;

   void VBpLayer::setDh(const ValVector & dhvec) { 
      dh().copyFrom(dhvec) ; 
   } ;
   void VBpLayer::getDh(ValVector & dhvec) const { 
      dh().copyTo(dhvec) ; 
   } ;

   //----------------------------------------------------------------------
   // fix equiv

   Bool VBpLayer::fixEquiv(const VBpLayerTable & layerTable) {
      if(isNull(thrEquivInfo)) return False ;

      StrForm * objinfo = thrEquivInfo->assoc(LblObj) ;
      if(isNull(objinfo)) {
	 ITK_ERR("Illegal StrForm for equiv info:" << *thrEquivInfo) ;
	 throw 0 ;
      } 

      // Here, objinfo should be "#layer[...]"

      SString layername ; 
      objinfo->assocCvtSString(LblName,layername,False) ;
      thrEquivLayer = layerTable.find(layername) ;
      if(isNull(thrEquivLayer)) {
	 ITK_ERR("Unknown layer name:" << layername) ;
	 throw 0 ;
      }
      
      SString vecname ;
      thrEquivInfo->assocCvtSString(LblVec,vecname,False) ;
      _thr = &(thrEquivLayer->vec(vecname)) ;

      return True ;
   } ;

   //======================================================================
   // I/O

   //----------------------------------------------------------------------
   // save

   void VBpLayer::save(ostream & ostr,
		       SaveLevel slevel,
		       const Bool & newlinep,
		       const SubString & indent) const {
      SubString indent2 = indent + "  " ; // for each param
      //SubString indent3 = indent2 + "  " ; // for each inner strForm

      StrForm::writeBeginTerm(ostr,LblTermHead) ;

      StrForm::writeAssocRole(ostr,LblName) ; 
      StrForm::writeString(ostr,name) ;
      StrForm::writeArgSep(ostr) ;

      StrForm::writeAssocRole(ostr,LblSize) ; 
      ostr << n << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,LblFuncType) ; 
      ostr << ftypeName(ftype) << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,LblLrnRate) ; 
      ostr << NNN_VAL_FORMAT << lrate << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,LblErrRate) ; 
      ostr << NNN_VAL_FORMAT2 << erate << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,LblSatRate) ; 
      ostr << NNN_VAL_FORMAT << satrate << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,LblSatVal) ; 
      ostr << NNN_VAL_FORMAT << satval << ITK_SST_TERM_SEP ;
      
      if(inpnorm > 0.0) {
	 StrForm::writeAssocRole(ostr,LblInpNorm) ;
	 ostr << NNN_VAL_FORMAT << inpnorm << ITK_SST_TERM_SEP ;
      }

      // save threshold
      ostr << endl << indent2 ;
      StrForm::writeAssocRole(ostr,LblThr) ; {
	 if(isNull(thrEquivInfo)) {
	    _thr->save(ostr,indent2 + "  ") ;
	 } else {
	    //ostr << thrEquivInfo->str ;
	    ostr << *thrEquivInfo ;
	 }
      } ;

      // save netinput and output
      if(slevel == SL_RunState || slevel == SL_AllState) { 
	 ostr << ITK_SST_TERM_SEP << endl << indent2 ;
	 StrForm::writeAssocRole(ostr,LblInp) ; {
	    _inp->save(ostr,indent2 + "  ") ;
	 } ;
	 ostr << ITK_SST_TERM_SEP << endl << indent2 ;
	 StrForm::writeAssocRole(ostr,LblOut) ; {
	    _out->save(ostr,indent2 + "  ") ;
	 } ;
      }

      // save other status
      if(slevel == SL_AllState) { 
	 ostr << ITK_SST_TERM_SEP << endl << indent2 ;
	 StrForm::writeAssocRole(ostr,LblErr) ; {
	    _err->save(ostr,indent2 + "  ") ;
	 } ;
	 ostr << ITK_SST_TERM_SEP << endl << indent2 ;
	 StrForm::writeAssocRole(ostr,LblDlt) ; {
	    _dlt->save(ostr,indent2 + "  ") ;
	 } ;
	 ostr << ITK_SST_TERM_SEP << endl << indent2 ;
	 StrForm::writeAssocRole(ostr,LblDh) ; {
	    _dh->save(ostr,indent2 + "  ") ;
	 } ;
      }

      StrForm::writeEndTerm(ostr) ;
      if(newlinep) ostr << endl ;
   } ;
   
   //----------------------------------------------------------------------
   // load

   void VBpLayer::scan(StrForm & sform, VBpLayerTable * layertable) {
      if(!sform.isTerm(LblTermHead)) {
	 ITK_ERR("Illegal VBpLayer StrForm:" << sform) ;
	 throw 0 ;
      } 

      name = sform.assoc(LblName)->cvtSString() ;

      ftype = ftypeByName(sform.assoc(LblFuncType)->cvtSString()) ;

      resize(sform.assoc(LblSize)->cvtUInt()) ;

      sform.assocCvtFlt(LblLrnRate,lrate) ;
      sform.assocCvtFlt(LblErrRate,erate) ;

      sform.assocCvtFlt(LblSatRate,satrate) ;
      sform.assocCvtFlt(LblSatVal,satval) ;
      sform.assocCvtFlt(LblInpNorm,inpnorm) ;

      initInfo = ITK_NULLPTR ;
      thrEquivInfo = ITK_NULLPTR ;
      StrForm * thrinfo = sform.assoc(LblThr) ;
            
      if(isNull(thrinfo)) {
	 /* ??? */
      } else if(thrinfo->isTerm(ValVector::LblTermHead)) { /* ValVector */
	 _thr->scan(*thrinfo) ;
      } else if(thrinfo->isTerm(LblEquiv)) { /* equiv */
	 thrEquivInfo = thrinfo->dup() ;
	 if(!isNull(layertable)) {
	    fixEquiv(*layertable) ;
	 }
      } else if(thrinfo->isTerm(LblInit)) { /* init */
	 initInfo = thrinfo->dup() ;
	 if(initInfo->isTerm(LblRand)) {
	    randomizeThreshold(initInfo->nthListArg(0)->cvtFlt(),
			       initInfo->nthListArg(1)->cvtFlt()) ;
	 } else {
	    ITK_ERR("illegal init value specifier:" << *initInfo) ;
	    throw 0 ;
	 }
	 initInfo = ITK_NULLPTR ;
      }

   } ;

   //----------------------------------------------------------------------
   // duplicate

   VBpLayer & VBpLayer::duplicate(const Bool sharedp) const {
      VBpLayer * v = new VBpLayer(name,n,ftype,True) ;
      copyTo(*v,sharedp,True) ;
      return *v ;
   } ;

   //----------------------------------------------------------------------
   // copyFrom

   VBpLayer & VBpLayer::copyFrom(const VBpLayer & src, 
				 const Bool sharedp,
				 const Bool allparamp) {
      if(allparamp) {
	 name.copy(src.name) ;
	 resize(src.n,True) ;
	 ftype = src.ftype ;
	 lrate = src.lrate ;
	 erate = src.erate ;
	 satrate = src.satrate ;
	 satval = src.satval ;
	 inpnorm = src.inpnorm ;

	 initInfo = src.initInfo->dup() ;
	 thrEquivInfo = src.thrEquivInfo->dup() ;
      }

      if(sharedp) { 
	 _thr = src._thr ;
      } else { 
	 _thr->copyFrom(*src._thr) ;
      }

      _inp->copyFrom(*src._inp) ;
      _out->copyFrom(*src._out) ;
      _err->copyFrom(*src._err) ;
      _dlt->copyFrom(*src._dlt) ;
      _dh->copyFrom(*src._dh) ;
      
      return *this ;
   } ;
   
} ;
