// -*- Mode:C++ -*-
//Header:
//File: VBpNetBase.h
//Author: NODA, Itsuki
//Date: 2000/01/13
//

//ModifyHistory:
// 2000/01/13: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_VBpNetBase_h_
#define _Nnn_VBpNetBase_h_
////////////////////////////////////////////////////////////////////////

#include <fstream.h>
#include "n3btype.h"
#include "n3utility.h"
#include "StrForm.h"
#include "VBpLayer.h"
#include "VBpLLink.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // VBpNetBase

   class VBpNetBase : public WithDescriber {

	 //--------------------------------------------------
	 // for save/load/scan
      public:

	 virtual char* L_TermHead() const { return "VBpNet" ; } ;
	 static char* const L_Name = "name" ; 
	 static char* const L_Layer = "layer" ; 
	 static char* const L_LLink = "llink" ; 
	 static char* const L_Temp = "temp" ;
	 static char* const L_Eps = "eps" ;
	 static char* const L_Alpha = "alpha" ;
	 static char* const L_CntTrial = "cntTrial" ;
	 static char* const L_CntLearn = "cntLearn" ;

#define NNN_DFLT_TEMP	1.0
#define NNN_DFLT_EPS	0.01
#define NNN_DFLT_ALPHA	0.5

	 //--------------------------------------------------
	 // member
      public:
	 SString name ;
	 Val temp ;	// temperature of activation function
	 Val eps ;	// step of learning
	 Val alpha ;	// inertia param for learning
	 VBpLayerTable layerTable ;
	 VBpLLinkTable linkTable ;

	 UInt cntTrial ; // count of trial (learn + test run)
	 UInt cntLearn ; // count of learning 

	 //--------------------------------------------------
	 // constructor
      public:
	 VBpNetBase() { init() ; } ;
	 VBpNetBase(const SString & nm,
		   const Val & tempval = NNN_DFLT_TEMP,
		   const Val & epsval = NNN_DFLT_EPS,
		   const Val & alphaval = NNN_DFLT_ALPHA) { 
	    init(nm,tempval,epsval,alphaval) ; } ;

	 //--------------------------------------------------
	 // initialize
      public:
	 void init(const SString & nm = SString(),
		   const Val & tempval = NNN_DFLT_TEMP,
		   const Val & epsval = NNN_DFLT_EPS,
		   const Val & alphaval = NNN_DFLT_ALPHA) { 
	    name.copy(nm,True) ;
	    temp = tempval ;
	    eps = epsval ;
	    alpha = alphaval ;
	    cntTrial = 0 ;
	    cntLearn = 0 ;
	 } ;

	 //--------------------------------------------------
	 // new layer/link
      public:
	 VBpLayer * newLayer(StrForm & sform) ;
	 VBpLayer * newLayer(const SString & nm, UInt size = 0, 
			    VBpLayer::FuncType ftype = VBpLayer::FT_sigmoid) ;
	 VBpLayer * newLayer(const VBpLayer * src) ; 

	 VBpLLink * newLLink(StrForm & sform) ;
	 VBpLLink * newLLink(VBpLayer * from, VBpLayer * to) ;
	 VBpLLink * newLLink(const VBpLLink * src, const Bool addp = True) ;

	 //--------------------------------------------------
	 // add layer/link
      public:
	 void addLayer(VBpLayer * layer) ;
	 void addLLink(VBpLLink * link) ;

	 //--------------------------------------------------
	 // access
      public:
	 VBpLayer * layer(const SString & nm) ;

	 VBpLLink * link(const SString & from, const SString & to) ;
	 VBpLLink * link(VBpLayer * from, VBpLayer * to) ;

	 VBpLayer * findLayer(const SString & nm) { return layer(nm) ; } ;
	 VBpLLink * findLink(const SString & from, const SString & to) {
	    return link(from,to) ;
	 } ;
	 VBpLLink * findLink(VBpLayer * from, VBpLayer * to) {
	    return link(from,to) ;
	 } ;

	 //--------------------------------------------------
	 // randomize threshold and weight
      public:
	 void randomize(const Val & minval, const Val & maxval) ;

	 //--------------------------------------------------
	 // save 
      public:
	 virtual void save(const SubString& file, 
			   SaveLevel slevel = SL_NetState) {
	    if(False) { // old version
	       ofstream ostr(file) ;
	       return save(ostr,slevel) ;
	    } else {    // more safety version
	       char * buffer = new char[file.length() + 8] ;
	       sprintf(buffer,"%s_XXXXXX%c",(const char*)file,ITK_NULLCHR) ;
	       char * tmpfile = mktemp(buffer) ;
	       ofstream ostr(tmpfile) ;
	       save(ostr,slevel) ;
	       ostr.close() ;
	       rename(tmpfile,file) ;
	    }
	 } ;
	 virtual void save(ostream & ostr, 
			   SaveLevel slevel = SL_NetState,
			   const Bool & newlinep = True,
			   const SubString & indent = SubString("")) ;

	 //--------------------------------------------------
	 // load
      public:
	 virtual void load(const SubString & file) {
	    ifstream istr(file) ;
	    if(!istr) {
	       ITK_ERR("can not open file \"" << file << "\".") ;
	       throw 0 ;
	    }
	    return load(istr) ;
	 } ;
	 virtual void load(istream & istr){
 	    StrForm * sform = StrForm::scan(istr) ;
	    scan(*sform) ;
	    delete sform ;
	 } ;


	 //--------------------------------------------------
	 // scan
      public:
	 virtual void scan(StrForm & sform) ;

	 //--------------------------------------------------
	 // duplicate 
      public:
	 VBpNetBase & duplicate(const Bool sharedp = False) const ;

	 //--------------------------------------------------
	 // copy
      public:
	 VBpNetBase & copyTo(VBpNetBase & dest, const Bool sharedp = False, 
			     const Bool allparamp = False) const {
	    return dest.copyFrom(*this, sharedp, allparamp) ;
	 } ;
	 VBpNetBase & copyFrom(const VBpNetBase & src, 
			       const Bool sharedp = False,
			       const Bool allparamp = False) ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#" << L_TermHead() << "[" << name << "]" ;

	    if(detailp) {
	       ostr << endl ;
	       ostr << "  " << "temp=" << NNN_VAL_FORMAT << temp << endl ;
	       ostr << "  " << "eps=" << NNN_VAL_FORMAT << eps << endl ;
	       ostr << "  " << "alpha=" << NNN_VAL_FORMAT << alpha << endl ;
	       ostr << "  " << "cntTrial=" << cntTrial << endl ;
	       ostr << "  " << "cntLearn=" << cntLearn << endl ;
	       ostr << "  " << "layer=" ;  layerTable.describe(ostr,True) ;
	       ostr << "  " << "llink=" ;  linkTable.describe(ostr,True) ;
	    } 
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif


