// -*- Mode:C++ -*-
//Header:
//File: VElman5L.h
//Title: VElman Net with 5 sigmoid layers
//Author: NODA, Itsuki
//Date: 2000/01/05
//

//ModifyHistory:
// 2000/01/05: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_VElman5L_h_
#define _Nnn_VElman5L_h_
////////////////////////////////////////////////////////////////////////

#include "VBpNetBase.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // VElman 5 Layer

   class VElman5L : public VBpNetBase {

      public:
	 //--------------------------------------------------
	 // constants

	 static char * const nameContext = "context" ;
	 static char * const nameInput   = "input" ;
	 static char * const nameInter1  = "inter1" ;
	 static char * const nameHidden  = "hidden" ;
	 static char * const nameInter2  = "inter2" ;
	 static char * const nameOutput  = "output" ;

	 //--------------------------------------------------
	 // member
      public:
	 VBpLayer * context ;
	 VBpLayer * input ;
	 VBpLayer * inter1 ;
	 VBpLayer * hidden ;
	 VBpLayer * inter2 ;
	 VBpLayer * output ;

	 VBpLLink * con_int1 ;
	 VBpLLink * inp_int1 ;
	 VBpLLink * int1_hid ;
	 VBpLLink * hid_int2 ;
	 VBpLLink * int2_out ;

	 //--------------------------------------------------
	 // constructor
      public:
	 VElman5L() { init() ; } ;

	 //--------------------------------------------------
	 // alloc new memory
     public:
	 void alloc(const SString & nm,
		    const UInt & inpsz,
		    const UInt & int1sz,
		    const UInt & hidsz,
		    const UInt & int2sz,
		    const UInt & outsz,
		    const Val & tempval = NNN_DFLT_TEMP,
		    const Val & epsval = NNN_DFLT_EPS,
		    const Val & alphaval = NNN_DFLT_ALPHA) {
	    init(nm,tempval,epsval,alphaval) ;
	    assignLayerLLink(inpsz,int1sz,hidsz,int2sz,outsz) ;
	 } ;
	 
	 //--------------------------------------------------
	 // assignLayer (l-s-l-s-l)
      public:
	 void assignLayerLLink(const UInt & inpsz, 
			       const UInt & int1sz,
			       const UInt & hidsz, 
			       const UInt & int2sz,
			       const UInt & outsz) {

	    context = newLayer(nameContext, hidsz,  VBpLayer::FT_linear) ;
	    input   = newLayer(nameInput,   inpsz,  VBpLayer::FT_linear) ;
	    inter1  = newLayer(nameInter1,  int1sz, VBpLayer::FT_sigmoid) ;
	    hidden  = newLayer(nameHidden,  hidsz,  VBpLayer::FT_linear) ;
	    inter2  = newLayer(nameInter2,  int2sz, VBpLayer::FT_sigmoid) ;
	    output  = newLayer(nameOutput,  outsz,  VBpLayer::FT_linear) ;
	    
	    con_int1 = newLLink(context, inter1) ;
	    inp_int1 = newLLink(input,   inter1) ;
	    int1_hid = newLLink(inter1,  hidden);
	    hid_int2 = newLLink(hidden,  inter2) ;
	    int2_out = newLLink(inter2,  output) ;
	 } ;

	 //--------------------------------------------------
	 // propFore
      public:
	 void propFore(const ValVector & inpvec) {
	    context->setOut(hidden->out()) ;

	    input->propFore(inpvec,temp) ;
	    
	    inter1->propFore(temp) ;
	    hidden->propFore(temp) ;
	    inter2->propFore(temp) ;
	    output->propFore(temp) ;
	 } ;
	    
	 //--------------------------------------------------
	 // propBack
      public:
	 void propBack(const ValVector & tchvec) {
	    output->propBack(tchvec,temp) ;
	    inter2->propBack(temp) ;
	    hidden->propBack(temp) ;
	    inter1->propBack(temp) ;
	 } ;

	 //--------------------------------------------------
	 // learn
      public:
	 Val learn() {
	    Val dsum = 0.0 ;

	    dsum += inter1->learn(eps,alpha) ;
	    dsum += hidden->learn(eps,alpha) ;
	    dsum += inter2->learn(eps,alpha) ;
	    dsum += output->learn(eps,alpha) ;

	    dsum += con_int1->learn(eps,alpha) ;
	    dsum += inp_int1->learn(eps,alpha) ;
	    dsum += int1_hid->learn(eps,alpha) ;
	    dsum += hid_int2->learn(eps,alpha) ;
	    dsum += int2_out->learn(eps,alpha) ;
	    
	    cntLearn++ ;
	    
	    return dsum ;
	 } ;

	 //--------------------------------------------------
	 // run
      public:
	 void run(const ValVector & inpvec) { 
	    propFore(inpvec) ; 
	    cntTrial++ ;
	 } ;
	 Val runLearn(const ValVector & inpvec,
		       const ValVector & tchvec) {
	    propFore(inpvec) ;
	    cntTrial++ ;
	    propBack(tchvec) ;
	    return learn() ;
	 } ;

	 //--------------------------------------------------
	 // get sum of square error
      public:
	 Flt sumError() const { return output->sumError() ; } ;

	 //--------------------------------------------------
	 // get output
      public:
	 void getNetOutput(ValVector & outvec) const { 
	    output->getNetOutput(outvec) ;
	 } ;

	 //--------------------------------------------------
	 // scan
      public:
	 virtual void scan(StrForm & sform) {
	    VBpNetBase::scan(sform) ;

	    context = layer(nameContext) ;
	    input   = layer(nameInput) ;
	    inter1  = layer(nameInter1) ;
	    hidden  = layer(nameHidden) ;
	    inter2  = layer(nameInter2) ;
	    output  = layer(nameOutput) ;

	    con_int1 = link(context, inter1) ;
	    inp_int1 = link(input,   inter1) ;
	    int1_hid = link(inter1,  hidden) ;
	    hid_int2 = link(hidden,  inter2) ;
	    int2_out = link(inter2,  output) ;
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif


