## -*- Mode: Ruby -*-

require 'gtk' ;

module Gtk

  class NamedColor
    attr :name		,true ;
    attr :r		,true ;
    attr :g		,true ;
    attr :b		,true ;

    RGBTXTFile = "/usr/X11R6/lib/X11/rgb.txt" ;

    def initialize(name)
      @name = nil ;

      rgbtxt = open(RGBTXTFile) ;
      while (line = rgbtxt.gets)
	next if line =~ /^\s*\!/ ;
	line.chop ;
	/\s*([0-9]+)\s+([0-9]+)\s+([0-9]+)\s+([^\s].*[^\s])/ =~ line ;
	if($4 == name) 
	  @r = $1.to_i * 0x100 ; 
	  @g = $2.to_i * 0x100 ; 
	  @b = $3.to_i * 0x100 ; 
	  @name = $4 ;
	  return ;
	end
      end
      raise "unknown color:" + name ;
    end
  end


  ##======================================================================
  ## class BufferedDrawingArea

  class BufferedDrawingArea < DrawingArea

    attr :buffer	,true ;
    attr :sizeX		,true ;
    attr :sizeY		,true ;
    attr :fgc		,true ;
    attr :bgc		,true ;
    attr :whiteGC	,true ;
    attr :blackGC	,true ;

    ##------------------------------------------------------------
    ## initialize

    def initialize(xsize = 0, ysize = 0)
      super() ;
      signal_connect("expose_event") {|w,e| exposeEvent(w,e) }
      set_usize(xsize,ysize) ;
    end

    ##------------------------------------------------------------
    ## set usize

    def set_usize(xsize,ysize)
      @xsize = xsize ;
      @ysize = ysize ;
      
      super(xsize,ysize) ;
    end

    ##------------------------------------------------------------
    ## newColoredGC

    def newColoredGC(colorspec)
      if(colorspec.is_a?(String))
	return newColoredGCbyName(colorspec)
      elsif (colorspec.is_a?(Array))
	return newColoredGCbyRGB(*colorspec)
      else
	raise "unknown colorspec format" + colorspec.to_s ;
      end
    end

    ##------------------------------------------------------------
    ## newColoredGCbyRGB

    def newColoredGCbyRGB(r,g,b)
      color = Gdk::Color.new(r,g,b) ;
      @colormap.alloc_color(color,TRUE,TRUE) ;
      gc = Gdk::GC::new(window) ;
      gc.copy(@bgc) ;
      gc.set_foreground(color) ;
      #    gc.set_background(color) ;
      gc ;
    end

    ##------------------------------------------------------------
    ## newColoredGCbyName

    def newColoredGCbyName(name) 
      cval = NamedColor::new(name) ;
      if(cval.name == nil) 
	return 
      end

      color = Gdk::Color.new(cval.r,cval.g,cval.b) ;
      @colormap.alloc_color(color,TRUE,TRUE) ;
      gc = Gdk::GC::new(window) ;
      gc.copy(@bgc) ;
      gc.set_foreground(color) ;
      #    gc.set_background(color) ;
      gc ;
    end

    ##------------------------------------------------------------
    ## setup

    def setup()
      if(@buffer.nil?) 
	@buffer = Gdk::Pixmap::new(window, @xsize, @ysize, -1) ;
	@colormap = Gdk::Colormap.get_system() ;
	@bgc = self.style.bg_gc(self.state) ;
	@fgc = self.style.fg_gc(self.state) ;
	@blackGC = newColoredGCbyRGB(0x0000,0x0000,0x0000) ;
	@whiteGC = newColoredGCbyRGB(0xFFFF,0xFFFF,0xFFFF) ;
	return TRUE ;
      else
	return FALSE ;
      end
    end

    ##------------------------------------------------------------
    ## expose Event

    def exposeEvent(w,e) 
      setup() if (@buffer.nil?) ;

      w.window.draw_pixmap(@bgc,
			   @buffer,e.area.x,e.area.y,
			   e.area.x,e.area.y,e.area.width,e.area.height) ;
    end

    ##------------------------------------------------------------
    ## expose all

    def exposeAll()
      draw(Gdk::Rectangle.new(0,0,@xsize,@ysize)) ;
    end

    ##------------------------------------------------------------
    ## flush

    def flush()
      exposeAll()
    end

    ##------------------------------------------------------------
    ## clear

    def clear(gc = @bgc)
      setup() ;
      @buffer.draw_rectangle(gc,TRUE,0,0,@xsize,@ysize) ;
    end

    ##------------------------------------------------------------
    ## draws

    def draw_point(gc,x,y)
      @buffer.draw_point(gc,x,y) ;
    end

    def draw_line(gc,x0,y0,x1,y1)
      @buffer.draw_line(gc,x0,y0,x1,y1) ;
    end

    def draw_lines(gc,points) ## points ::= [[x0,y0],[x1,y1],...]
      @buffer.draw_lines(gc,points) ;
    end

    def draw_rectangle(gc,fill,x,y,w,h)
      @buffer.draw_rectangle(gc,fill,x,y,w,h) ;
    end

    def draw_arc(gc,fill,x,y,w,h,a1,a2)
      @buffer.draw_arc(gc,fill,x,y,w,h,a1,a2) ;
    end

    def draw_rectangle_centered(gc,fill,x,y,w,h)
      draw_rectangle(gc,fill,x-w/2,y-h/2,w,h) ;
    end

    def draw_circle_centered(gc,fill,x,y,r)
      draw_arc(gc,fill,x-r/2,y-r/2,r,r,0,64*360) ;
    end

  end

  ##======================================================================
  ## class BufferedDrawingAreaWithBack 

  class BufferedDrawingAreaWithBack < BufferedDrawingArea
    attr :back		,true ;
    
    def setup()
      if(super())
	@back = Gdk::Pixmap::new(window, @xsize, @ysize, -1) ;
	@buffer = Gdk::Pixmap::new(window, @xsize, @ysize, -1) ;
      end
    end

    def saveToBack()
      @back.draw_pixmap(@bgc,@buffer,0,0,0,0,@xsize,@ysize) ;
    end

    def restoreFromBack()
      @buffer.draw_pixmap(@bgc,@back,0,0,0,0,@xsize,@ysize) ;
    end
      
  end

end
