// -*- Mode:C++ -*-
//Header:
//File: SimplePlayer.cc
//Title: Base Player for Soccer Server 7
//Author: NODA, Itsuki
//Date: 2001/03/10
//

//ModifyHistory:
// 2001/03/10: Start to create this file
//EndModifyHistory:

#include "SimplePlayer.h"

namespace Saloo {

   //======================================================================
   // state manipuration

   FieldState & SimplePlayer::pushState(const CycleStep & time, 
					const FieldState & s) {
      FieldState & fs = stateEntry(time.count) ;
      fs.copyFrom(s) ;
      senseStateTime = time ;
      estimateStateTime = time ;

      /* <<< [I.Noda:2001.07.26] <<< */
      worldmodel.putSensedData(time.count,s) ;
      /* >>> [I.Noda:2001.07.26] >>> */

      return fs ;
   }

   //======================================================================
   // analyze see info

   Bool SimplePlayer::analyzeSensorSee() { 
      currentState().estimate(*this) ;
      pushState(vtime(),currentState()) ;
      if(stime() <= vtime()) { stime() = vtime() ; } ;
      return True ; 
   } ;

   //======================================================================
   // analyze hear info

   Bool SimplePlayer::analyzeSensorHear() { 
      if(sinfo.hear.sender == HearInfo::Referee) {
	 cvPlayMode.lock() ; {
	    playmode() = sinfo.hear.aux.pmode ;
	    cvPlayMode.broadcast() ;
	    //ITK_DBG(*playmode()) ;
	 } ; cvPlayMode.unlock() ;
      }
      atime() = sinfo.time ;
      if(stime() <= atime()) { stime() = atime() ; } ;
      return True ;
   } ;

   //======================================================================
   // analyze sensebody info

   Bool SimplePlayer::analyzeSensorSenseBody() { 
      myneck() = sinfo.senseBody.headAngle ;
      btime() = sinfo.time ;
      stime() = btime() ;
      return True ;
   } ;

   //======================================================================
   // analyze see info

   Bool SimplePlayer::analyzeSensorFullState() { 

      taughtState().time() = sinfo.fullstate.time ;
      taughtState().ballpos() = sinfo.fullstate.ballpos ;
      taughtState().ballvel() = sinfo.fullstate.ballvel ;
      FullStateInfo::Player * p = ITK_NULLPTR;
      for(UInt i = 0 ; i < sinfo.fullstate.playerN ; i++) {
	 if(sinfo.fullstate.player[i].side == side &&
	    sinfo.fullstate.player[i].unum == unum) {
	    p = &(sinfo.fullstate.player[i]) ;
	    break;
	 }
      }
      taughtState().myposdir() = p->posdir ;
      taughtState().myvel() = p->vel ;
      
      fstime() = sinfo.time ;
      stime() = fstime() ;
      return True ;
   } ;

   //======================================================================
   // Command Timing

   Bool SimplePlayer::waitNextBestComTime(UInt marginMS) {
      if(bctime() >= btime()) cvSenseBody.lockWait() ;
      TimeVal timeup = btime().rtime + TimeVal(0,100-marginMS) ;
      //ITK_VBS(1,"wait :" << timeup) ;
      sleepUntil(timeup) ;
      return True ;
   } 

   //======================================================================
   // Wrapper for Super Class

   Bool SimplePlayer::sendCommandPost(Bool bodyp) {
      ctime().set(stime().count,True) ;
      if(bodyp) bctime().set(stime().count,True) ;
      return ThreadedPlayer::sendCommandPost(bodyp) ;
   } ;

   //======================================================================
   // Basic Command

   //--------------------------------------------------
   // move

   Bool SimplePlayer::move(const Flt & x, const Flt & y) { 
      return sendMove(x,y) ; 
   } ;

   //--------------------------------------------------
   // turn

   const Flt TurnCutOffAngle = 1.0 ;

   Bool SimplePlayer::turn(const Flt & power_) { 
      Flt power = normalizeAngle(power_) ;
      if(Abs(power) < TurnCutOffAngle) return False ;

      Flt inertia_moment = 5.0 ; /* ***** from server.conf */
      Flt realpower = 
	 power * (1.0 + inertia_moment * myspeed()) ;

      Bool r = sendTurn(realpower) ; 

      if(r) {
	 mybdir() += power ;
	 myhdir() += power ;
      }	 
      
      return r ;
   } ;

   //--------------------------------------------------
   // dash

   Bool SimplePlayer::dash(const Flt & power) { 
      return sendDash(power) ; 
   } ;

   //--------------------------------------------------
   // kick

   Bool SimplePlayer::kick(const Flt & power, const Flt & dir_) {
      Flt dir = normalizeAngle(dir_) ;
      return sendKick(power,dir) ;
   } ;

   //--------------------------------------------------
   // goalie catch

   Bool SimplePlayer::gcatch(const Flt & dir) {
      return sendCatch(dir) ;
   } ;

   //--------------------------------------------------
   // turn neck

   Bool SimplePlayer::head(const Flt & dir) {
      Flt neck = normalizeAngle(myneck() + dir) ;
      if(neck >  90.0) neck =  90.0 ;
      if(neck < -90.0) neck = -90.0 ;

      Flt moment = normalizeAngle(neck - myneck()) ;

      Bool r = sendTurnNeck(moment) ;
      
      if(r) {
	 myneck() = neck ;
	 myhdir() = mybdir() + myneck() ;
      }

      return r ;
   } ;

   //--------------------------------------------------
   // say

   Bool SimplePlayer::say(const SubString & msg) {
      return sendSay(msg) ;
   } ;

   //======================================================================
   // low level command

   //--------------------------------------------------
   // turns

   Bool SimplePlayer::turnTo(const Flt adir) {
      return turn(normalizeAngle(adir - mybdir())) ;
   } ;

   Bool SimplePlayer::turnToRel(const Vec & rpos) {
      return turnTo(rpos.angle()) ;
   } ;

   Bool SimplePlayer::turnToAbs(const Vec & apos) {
      return turnToRel(mypos().diffTo(apos)) ;
   } ;

   //--------------------------------------------------
   // kicks

   Bool SimplePlayer::kickTo(const Flt & power, const Flt & adir) {
      return kick(power, normalizeAngle(adir - mybdir())) ;
   } ;

   Bool SimplePlayer::kickToRel(const Flt & power, const Vec & rpos) {
      return kickTo(power, rpos.angle()) ;
   } ;

   Bool SimplePlayer::kickToAbs(const Flt & power, const Vec & apos) {
      return kickToRel(power, mypos().diffTo(apos)) ;
   } ;

   //--------------------------------------------------
   // gcatch

   Bool SimplePlayer::gcatchTo(const Flt & adir) {
      return gcatch(normalizeAngle(adir - mybdir())) ;
   } ;

   Bool SimplePlayer::gcatchToRel(const Vec & rpos) {
      return gcatchTo(rpos.angle()) ;
   } ;

   Bool SimplePlayer::gcatchToAbs(const Vec & apos) {
      return gcatchToRel(mypos().diffTo(apos)) ;
   } ;

   //--------------------------------------------------
   // neck

   Bool SimplePlayer::headTo(const Angle & adir) {
      return head(normalizeAngle(adir - myhdir())) ;
   } ;

   Bool SimplePlayer::headToRel(const Vec & rpos) {
      return headTo(rpos.angle()) ;
   } ;

   Bool SimplePlayer::headToAbs(const Vec & apos) {
      return headToRel(mypos().diffTo(apos)) ;
   } ;

} ;
   
 
   

